# coding: utf-8
"""
    Baseline.Convex.py

    Copyright (c) 2020-2025, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.spatial import ConvexHull
from scipy.optimize import minimize
from matplotlib.gridspec import GridSpec
from molass_legacy.DataStructure.MatrixData import simple_plot_3d

TITLE_FONTSIZE = 16

def get_baseline_vertices(size, vertices):
    n0 = np.argmin(vertices)
    n1 = np.argmax(vertices)
    assert vertices[n0] == 0
    assert vertices[n1] == size - 1

    if n1 > n0:
        return vertices[n0:n1+1]
    else:
        return np.hstack([vertices[n0:], vertices[0:n1+1]])

def compute_baseline(y, vertices):
    i = vertices[0]
    yi = y[i]
    ret_y = np.zeros(len(y))

    for j in vertices[1:]:
        w = j-i
        w_ = np.arange(w)/w
        yj = y[j]
        dy = yj - yi
        ret_y[i:j] = yi + dy*w_
        i = j
        yi = yj

    ret_y[i] = yi
    return ret_y

def get_best_adjust_params(y, ys, end_limits=(10, -10)):
    """
    it is better to change end_limits to suite the situation
    """

    zeros = np.zeros(len(y))
    init_rate = y[-1]/ys[-1]
    ls = slice(0, end_limits[0])
    ms = slice(end_limits[0], end_limits[1])
    rs = slice(end_limits[1], None)

    def obj_func(x):
        rate, base = x
        return (np.sum((y[ls] - (base + ys[ls]*rate))**2)
                    + np.sum(np.min([y[ms] - (base + ys[ms]*rate), zeros[ms]], axis=0)**2)
                    + np.sum((y[rs] -(base + ys[rs]*rate))**2)
                    )

    res = minimize(obj_func, (init_rate, 0))
    return res.x

def demo(root, in_folder, xray=True, show_preview=False, debug=False):
    import molass_legacy.KekLib.DebugPlot as plt
    from MeasuredData import MeasuredData

    print(in_folder)
    md = MeasuredData(in_folder)

    if show_preview:
        from molass_legacy.Tools.ThreeDimViewer import ThreeDimViewer
        dialog = ThreeDimViewer(root, md)
        dialog.show()

    if xray:
        rd = md.xr
        ecolor = 'orange'
    else:
        rd = md.uv
        ecolor = 'blue'

    rd.set_elution_curve()
    i_slice = rd.i_slice
    j_slice = rd.j_slice

    whole_data = rd.data
    data = whole_data[i_slice,j_slice]
    q = rd.vector[i_slice]

    fig = plt.figure(figsize=(21, 11))
    gs = GridSpec(3,4)
    ax00 = fig.add_subplot(gs[0,0], projection='3d')

    axes = []
    for i in range(3):
        for j in range(4):
            if i+j == 0:
                pass
            else:
                axes.append(fig.add_subplot(gs[i,j]))

    ax00.set_title("3D Data View", y=1.09, fontsize=TITLE_FONTSIZE)
    simple_plot_3d(ax00, data, x=q)
    z = rd.e_curve.y[j_slice]

    q1 = rd.vector[rd.e_index]
    size = data.shape[1]
    x = np.ones(size)*q1
    y = np.arange(size)

    ax00.plot(x, y, z, color=ecolor)

    baseline = None
    zeros = np.zeros(size)
    for ax in axes:
        ax.plot(y, z, color=ecolor)

        if baseline is None:
            points = np.array(list(zip(y,z)))
            hull = ConvexHull(points)
            ax.plot(points[hull.vertices,0], points[hull.vertices,1], 'o-', color='yellow')

            print('hull.vertices=', hull.vertices)
            vertices = get_baseline_vertices(size, hull.vertices)
            ax.plot(points[vertices,0], points[vertices,1], 'o', color='red')
            baseline = compute_baseline(z, vertices)

        ax.plot(y, baseline, ':', color="red")

        zc = np.max([z - baseline, zeros], axis=0)
        # zc = z - baseline
        zs = np.cumsum(zc)
        rate, base = get_best_adjust_params(z, zs)

        if debug:
            ax.plot(y, zs*0.01, color='cyan')
            print('rate, base=', rate, base)

        zs_ = zs*rate + base
        ax.plot(y, zs_, '-', color="red")
        baseline = zs_

    fig.tight_layout()
    plt.show()

def integrative_curve_convex(y, baseline=None, epsilon=1e-4, return_hull=False):
    size = len(y)

    if baseline is None:
        x = np.arange(size)
        points = np.array(list(zip(x,y)))
        hull = ConvexHull(points)
        vertices = get_baseline_vertices(size, hull.vertices)
        baseline = compute_baseline(y, vertices)
        if False:
            print('hull.vertices=', hull.vertices)
            plt.push()
            fig, ax = plt.subplots()
            ax.plot(x, y)
            ax.plot(points[hull.vertices,0], points[hull.vertices,1], 'o-', color='yellow')
            ax.plot(x, baseline, ':', color='red')
            fig.tight_layout()
            plt.show()
            plt.pop()

    min_e = None
    min_baseline = None
    zeros = np.zeros(size)
    for k in range(5):
        y_ = np.max([y - baseline, zeros], axis=0)
        # y_ = y - baseline
        cy = np.cumsum(y_)
        height = cy[-1] - cy[0]

        rate, base = get_best_adjust_params(y, cy)
        new_baseline = cy*rate + base

        e = np.max(np.abs(new_baseline - baseline))/abs(height)
        # print([k], 'e=', e)
        if e < epsilon:
            break

        if min_e is None or e < min_e:
            min_e = e
            min_baseline = new_baseline
        else:
            new_baseline = min_baseline
            break

        baseline = new_baseline

    if return_hull:
        return new_baseline, points[hull.vertices, :]
    else:
        return new_baseline

def proof_plot(in_folder, xray=True):
    import molass_legacy.KekLib.DebugPlot as plt
    from MeasuredData import MeasuredData
    from DataUtils import get_in_folder

    md = MeasuredData(in_folder)

    if xray:
        rd = md.xr
        ecolor = 'orange'
    else:
        rd = md.uv
        ecolor = 'blue'

    j_slice = rd.j_slice
    x = rd.e_curve.x[j_slice]
    y = rd.e_curve.y[j_slice]

    b, vertices = integrative_curve_convex(y, return_hull=True)

    j0 = j_slice.start
    if j0 is None:
        j0 = 0
    fig, ax = plt.subplots()
    ax.set_title("Proof Plot of Convex Hull Method for " + get_in_folder(in_folder), fontsize=20)
    ax.plot(x, y, color=ecolor, label='data')
    ax.plot(j0+vertices[:,0], vertices[:,1], 'o-', color='yellow', label='convex hull')
    ax.plot(x, b, ':', color='red', label='integral basecurve')

    ax.legend(fontsize=16)
    fig.tight_layout()
    plt.show()



"""
    SeccolFunctions.py

    Copyright (c) 2022-2025, SAXS Team, KEK-PF
"""
from molass_legacy.SecTheory.ColumnConstants import SECCONF_LOWER_BOUND

def rgfit_secconf_eval(rgfit, secconf):
    if secconf > SECCONF_LOWER_BOUND:
        return max(rgfit, secconf)
    else:
        return rgfit + SECCONF_LOWER_BOUND



# coding: utf-8
"""
    AbsorbanceViewUtils.py

    Copyright (c) 2018-2022, SAXS Team, KEK-PF
"""
import numpy                as np
from molass_legacy._MOLASS.SerialSettings         import get_setting
from .Absorbance import LOW_PERCENTILE_BASE_1ST

def draw_3d( ax, absorbance, wvlen_lower, wvlen_upper, i_slice, zlim=None,
                title="", title_fontsize=None, low_percentile=True ):
    ax.set_title( title, fontsize=title_fontsize )
    ax.set_xlabel( '\nwave length (nm)' )
    ax.set_ylabel( '\nsequential №' )
    ax.set_zlabel( '\nabsorbance' )

    if zlim is None:
        # zlim = ( -0.05, 0.2 )
        # ax.set_zlim( zlim )
        pass
    else:
        ax.set_zlim( zlim )

    data = absorbance.data
    print( 'data.shape=', data.shape )

    size = data.shape[1]

    peak_j = None
    tail_i = absorbance.tail_index
    absorbance_baseline_type = get_setting( 'absorbance_baseline_type' )

    if low_percentile:
        zpp = np.percentile( data, LOW_PERCENTILE_BASE_1ST )
        wpp = np.where( data < zpp )
        npp = data.shape[0]*data.shape[1]
        if False:
            print( 'data.shape=', data.shape, npp)
            print('wpp[0].shape=', wpp[0].shape, wpp[0].shape[0]/npp, 'wpp[1].shape=', wpp[1].shape[0], wpp[1].shape[0]/npp)
            print( 'wpp=', wpp )
        x_ = absorbance.wl_vector[wpp[0]]
        y_ = wpp[1]
        z_ = data[wpp[0], wpp[1]]
        ax.plot(x_, y_, z_, 'o', color='yellow')

    for i, wv in enumerate( absorbance.wl_vector ):
        if i == tail_i:
            pass
        else:
            if wv < wvlen_lower or wv > wvlen_upper:
                pass
            if i != absorbance.index and i % 2 > 0:
                continue

        X = np.ones( size ) * wv
        Y = np.arange( size )
        Z = data[i,:]
        if i == absorbance.index:
            alpha   = 1
            color   = 'blue'
        else:
            alpha   = 0.2
            color   = '#1f77b4'
        ax.plot( X, Y, Z, color=color, alpha=alpha )

        if absorbance_baseline_type == 0 or absorbance.base_curve is None:
            continue

        if i  == tail_i:
            slice_ = slice( absorbance.rail_points[0], absorbance.rail_points[1]+1 )
            Y_ = Y[slice_]
            curve_size = len(absorbance.tail_curve)
            for ii in range( i-5, i+6, 2 ):
                X_ = np.ones( curve_size ) * absorbance.wl_vector[ii]
                ax.plot( X_, Y_, absorbance.tail_curve, color='red', alpha=0.5 )

            X_ = np.ones( curve_size ) * absorbance.std_wvlen
            ax.plot( X_, Y_, absorbance.base_curve, color='red', alpha=0.5 )

    draw_3d_detail( ax, absorbance, i_slice )

def draw_3d_detail( ax, absorbance, i_slice ):
    wvlen   = absorbance.wl_vector
    absorbance  = absorbance

    wvlen_ = wvlen[i_slice]

    if False:
        lower_j  = absorbance.a_curve.lower_abs
        middle_j = absorbance.a_curve.middle_abs
        upper_j  = absorbance.a_curve.upper_abs
    else:
        n       = absorbance.a_curve.primary_peak_no
        info    = absorbance.a_curve.peak_info[n]
        lower_j  = info[0]
        middle_j = int(info[1] + 0.5)
        upper_j  = info[2]

    section_array = []
    # no flow_changes plot here
    colors = [ 'green', 'orange', 'green' ]
    for i, j in enumerate( [ lower_j, middle_j, upper_j ]):
        print( 'draw_3d_absorbance_detail: (i, j)=', (i, j) )
        if j is None:
            continue

        X = wvlen_
        Y = np.ones( len(wvlen_) ) * j
        Z = absorbance.data[i_slice,j]
        section_array.append( [ j, Z ] )
        color = colors[i]
        ax.plot( X, Y, Z, ':', color=color )

    bline = absorbance.get_bottomline_matrix()

    xmin, xmax = ax.get_xlim()
    zmin, zmax = ax.get_zlim()
    xoffset = ( xmax - xmin ) * 0.01
    zoffset = ( zmax - zmin ) * 0.005

    Y = np.arange( absorbance.data.shape[1] )
    for i in range(4):
        wvlen = absorbance.reg_wvlens[i]
        X = np.ones( absorbance.data.shape[1] ) * wvlen
        Z = bline[:, i]
        if i == 3:
            color   = 'cyan'
            alpha   = 1
        else:
            color   = 'red'
            alpha   = 0.5
        ax.plot( X, Y, Z, ':', color=color, alpha=alpha )
        if i in [0,3]:
            ax.text( wvlen - xoffset, Y[-1], Z[-1] + zoffset, 'λ=%g' % wvlen )

    if False:
        ipp = absorbance.istd + absorbance.wpp[0]
        xpp = absorbance.wl_vector[ ipp ]
        ypp = absorbance.jump_j_safe + absorbance.wpp[1]
        zpp = absorbance.data[ ipp, ypp ]
        ax.plot( xpp, ypp, zpp, 'o', color='yellow', markersize=3, alpha=0.1 )

        xff, yff, zff = absorbance.final_base_point
        print( 'xff, yff, zff=', xff, yff, zff )
        ax.plot( [xff], [yff], [zff], 'o', color='red', markersize=5 )



# coding: utf-8
"""

    RobustCurve.py

        evaluation of simularity between curves

    Copyright (c) 2021, SAXS Team, KEK-PF

"""
import numpy as np
from scipy.optimize import minimize
from molass_legacy.KekLib.SciPyCookbook import smooth
from matplotlib.gridspec import GridSpec
import molass_legacy.KekLib.DebugPlot as plt
from .RobustPeaks import RobustPeaks
from .ElutionModels import gaussian, egh

LIFTED_BASE = False
MINIMIZE_METHOD = 'Nelder-Mead'

class PeakModel:
    def get_peaktop(self):
        return self.peaktop

    def __call__(self, x=None):
        if x is None:
            x = self.px
        return x, self.base + self.func(x, *self.params)

class PeakGaussian(PeakModel):
    name = 'Gaussian'
    def __init__(self, x, y, pt, num_hw=10):
        pt_slice = slice(pt-num_hw, pt+num_hw)
        x_ = x[pt_slice]
        if LIFTED_BASE:
            b_ = np.min(y[pt_slice])
        else:
            b_ = 0
        y_ = y[pt_slice] - b_
        # y_ = smooth(sy, window_len=5)

        def obj_func(p):
            h, m, s = p
            return np.sum((gaussian(x_, h, m, s) - y_)**2)

        h_ = y_[num_hw]
        m_ = x[pt]
        s_ = (x_[-1] - x_[0])/2
        ret = minimize(obj_func, (h_, m_, s_), method=MINIMIZE_METHOD)
        h, m, s = ret.x
        self.px = x_ + m - x[pt]
        self.base = b_
        self.params = (h, m, s)
        self.peaktop = (m, b_+h)
        self.pt_slice = pt_slice
        self.func = gaussian

        if False:
            plt.push()
            fig, ax = plt.subplots()
            # ax.plot(x_, sy)
            ax.plot(x_, y_, ':')
            ax.plot(x_, gaussian(x_, h, m, s))
            plt.show()
            plt.pop()

class PeakEGH(PeakModel):
    name = 'EGH'
    def __init__(self, x, y, pt, num_hw=10):
        pt_slice = slice(pt-num_hw, pt+num_hw)
        x_ = x[pt_slice]
        if LIFTED_BASE:
            b_ = np.min(y[pt_slice])
        else:
            b_ = 0
        y_ = y[pt_slice] - b_

        self.debug = False

        def obj_func(p):
            h, m, s, t = p
            fy = egh(x_, h, m, s, t)
            if self.debug:
                plt.push()
                fig, ax = plt.subplots()
                ax.plot(x_, y_)
                ax.plot(x_, fy)
                self.debug = plt.show()
                plt.pop()
            return np.sum((fy - y_)**2)

        h_ = y_[num_hw]
        m_ = x[pt]
        s_ = (x_[-1] - x_[0])/2
        t_ = 0
        ret = minimize(obj_func, (h_, m_, s_, t_), method=MINIMIZE_METHOD)
        h, m, s, t = ret.x
        self.px = x_ + m - x[pt]
        self.base = b_
        self.params = (h, m, s, t)
        self.peaktop = (m, b_+h)
        self.pt_slice = pt_slice
        self.func = egh

class PeakHyperbola(PeakModel):
    name = 'Hyperbola'
    def __init__(self, x, y, pt, num_hw=10):
        from .Hyperbola import RotatedHyperbola, compute_rotated_hyperbola_y

        pt_slice = slice(pt-num_hw, pt+num_hw)
        x_ = x[pt_slice]
        if LIFTED_BASE:
            b_ = np.min(y[pt_slice])
        else:
            b_ = 0
        y_ = y[pt_slice] - b_

        ptx = x[pt]
        pty = y[pt] - b_
        hbw = (x_[-1] - x_[0])/2
        rh = RotatedHyperbola(ptx, pty, hbw=hbw)
        rh.fit(x_, y_, method=MINIMIZE_METHOD)

        self.base = b_
        self.params = rh.fitted_params
        self.peaktop = b_ + rh.get_peak_top()
        self.px = x_ + self.peaktop[0] - x[pt]
        self.pt_slice = pt_slice
        self.func = compute_rotated_hyperbola_y

        if False:
            plt.push()
            fig, ax = plt.subplots()
            # ax.plot(x_, sy)
            ax.plot(x_, y_, ':')
            ax.plot(x_, rh(x_))
            plt.show()
            plt.pop()

class RobustCurve:
    def __init__(self, x, y, peak_model=PeakEGH, debug=False):
        self.x = x
        self.y = y
        self.rp = RobustPeaks(x, y, debug=debug)
        self.peak_model = peak_model
        self.fit_peak_models()

    def fit_peak_models(self):
        pgs = []
        for peak in self.rp.get_peaks():
            ls, pt, rs = peak
            width = int(((pt - ls) + (rs - pt))/4)
            pgs.append(self.peak_model(self.x, self.y, pt, num_hw=width))
        self.peak_models = pgs

    def get_peak_tops(self):
        pts = []
        for pg in self.peak_models:
            pts.append(pg.get_peaktop())
        return pts

    def get_peak_models(self):
        return self.peak_models

def plot_curve(ax, curve, color=None):
    x = curve.x
    y = curve.y
    ax.plot(x, y, color=color)
    labeled = False
    for pg in curve.get_peak_models():
        g_label = None if labeled else 'locally fitted gaussian'
        ax.plot(*pg(), color='green', label=g_label)
        x_ = x[pg.pt_slice]
        y_ = pg.base
        ax.plot(x_[[0,-1]], [y_, y_], ':', color='red')
        pt_label = None if labeled else 'fitted gaussian top'
        ax.plot(*pg.get_peaktop(), 'o', color='red', label=pt_label)
        labeled = True

def demo(sd):
    xr_curve = sd.get_xray_curve()
    uv_curve = sd.get_uv_curve()

    uv_rc = RobustCurve(uv_curve.x, uv_curve.y)
    xr_rc = RobustCurve(xr_curve.x, xr_curve.y)

    plt.push()
    fig, (ax1, ax2, ax3) = plt.subplots(ncols=3, figsize=(21,7))
    ax1.set_title("UV Elution Curve", fontsize=16)
    ax2.set_title("Xray Elution Curve", fontsize=16)
    plot_curve(ax1, uv_rc)
    plot_curve(ax2, xr_rc)
    fig.tight_layout()
    plt.show()
    plt.pop()




"""
    Optimizer.ParameterTransition.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import os
import numpy as np
import queue
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy._MOLASS.SerialSettings import set_setting

USE_RESULT_PROXY = False

def show_parameter_transition_impl(selector_dialog, debug=True):
    if debug:
        from importlib import reload
        import molass_legacy.KekLib.OnTheFlyUtils
        reload(KekLib.OnTheFlyUtils)
        import Optimizer.ResultProxy
    from molass_legacy.KekLib.OnTheFlyUtils import show_progress
    from molass_legacy.Optimizer.ResultProxy import ResultProxy

    joblist_folder = selector_dialog.folder.get()
    print("show_parameter_transition_impl: joblist_folder=", joblist_folder)

    params_list = []
    bounds_list = []
    fv_list = []
    optimizer = None
    progress_queue = queue.Queue()

    num_steps = 100

    def prepare_data():
        nonlocal optimizer
        for i, node in enumerate(os.listdir(joblist_folder)):
            print([i], node)
            if i == num_steps:
                break

            jobpath = os.path.join(joblist_folder, node)
            set_setting('optworking_folder', jobpath)

            if USE_RESULT_PROXY:
                # this yet seems buggy
                if i == 0:
                    first_result = selector_dialog.get_result(folder=jobpath)
                    result = first_result
                else:
                    result = ResultProxy(jobpath, first_result)
            else:
                result = selector_dialog.get_result(folder=jobpath)

            optimizer = result.get_optimizer()
            bounds = optimizer.get_param_bounds(optimizer.init_params)
            bounds_list.append(bounds)
            for j, params in result.get_result_iterator(all=True):
                nparams = optimizer.to_norm_params(params)
                params_list.append(nparams)
            fv_list.append(result.fv_array)
            progress_queue.put(i+1)

    show_progress(prepare_data, progress_queue, num_steps)

    params_array = np.array(params_list)
    bounds_array = np.array(bounds_list)
    print("params_array.shape=", params_array.shape)

    def plot_transition():
        from importlib import reload
        import Optimizer.ParameterTransitionPlot as Plot
        reload(Plot)
        from molass_legacy.Optimizer.ParameterTransitionPlot import plot_transition_impl
        try:
            plot_transition_impl(optimizer, params_array, bounds_array, fv_list)
        except:
            import traceback
            traceback.print_exc()

    with plt.Dp(window_title="Parameter Transition",
                button_spec=["OK", "Cancel"],
                extra_button_specs=[("Plot", plot_transition)]):
        fig, ax = plt.subplots()
        fig.tight_layout()
        plt.show()


"""
    Peaks/PeakParamsSet.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np

class PeakParamsSet:
    def __init__(self, uv_peaks, xr_peaks, a, b):
        self.uv_peaks = np.asarray(uv_peaks)
        self.xr_peaks = np.asarray(xr_peaks)
        self.a = a
        self.b = b
        self.items = [self.uv_peaks, self.xr_peaks, a, b]

    def __iter__(self):
        # this is for backward compatibility
        return iter(self.items)

    def __getitem__(self, item):
        # this is for backward compatibility
        return self.items[item]



# coding: utf-8
"""
    RunAsAdmin.py
    Copyright (c) 2019, Masatsuyo Takahashi, KEK-PF
"""

"""
    borrowed from
    https://superuser.com/questions/615654/run-exe-file-via-python-as-administrator
"""
import ctypes, sys, os

def is_admin():
    try:
        return ctypes.windll.shell32.IsUserAnAdmin()
    except:
        return False

def runas():
    if is_admin():
        import time
        print("Running as admin.")
        time.sleep(3)
    else:
        # Re-run the program with admin rights
        import molass_legacy.KekLib.CustomMessageBox as MessageBox
        MessageBox.askokcancel("RunAsAdmin",
            'Please trust this program and reply "Yes" to the next dialog.'
            )
        ctypes.windll.shell32.ShellExecuteW(None, "runas", sys.executable, __file__, None, 1)

if __name__ == '__main__':
    runas()



# coding: utf-8
"""
    ErrorAnalysis.py

    Copyright (c) 2018, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.interpolate import LSQUnivariateSpline

NUM_KNOTS = 40

def smoothed_error( E, debug=False ):
    x = np.arange(E.shape[0])
    knots = np.linspace(x[0], x[-1], NUM_KNOTS)[1:-1]
    SE_list = []
    delta = 0.001
    for j in range(E.shape[1]):
        y = np.log( E[:,j] + delta )
        spline = LSQUnivariateSpline( x, y, knots )
        y_ = spline(x)
        if debug:
            if j == 85:
                debug_plot_smoothing( x, y, y_ )
        SE_list.append(np.exp(y_) - delta)
    return np.array(SE_list).T

def debug_plot_smoothing( x, y, y_ ):
    import molass_legacy.KekLib.DebugPlot as plt
    fig = plt.figure()
    ax = fig.gca()
    ax.plot(x, y, 'o', markersize=3)
    ax.plot(x, y_)
    fig.tight_layout()
    plt.show()

def compute_denoised_data( rank, X ):
    dslice = slice(0,rank)
    U, s, Vt = np.linalg.svd( X )
    X_ = np.dot(  np.dot( U[:,dslice], np.diag( s[dslice] ) ), Vt[dslice, :] )
    return X_

class ErrorAnalyzer:
    def __init__( self, M, E ):
        self.M = M
        self.E = smoothed_error(E)

    def compute_denoised_error( self, rank, num_iter ):
        M = self.M
        S   = np.zeros( M.shape )
        S2  = np.zeros( M.shape )
        S_  = np.zeros( M.shape )
        S_2 = np.zeros( M.shape )
        U_  = np.zeros( M.shape )
        U_2 = np.zeros( M.shape )

        M_ = compute_denoised_data( rank, M )
        Mpinv = np.linalg.pinv( M )
        W = np.dot( M_, Mpinv )

        for i in range(num_iter):
            if i % 100 == 0:
                print( 'iteration', i )
            E_ = self.generate_noise()
            X   = M + E_

            S   += X
            S2  += X**2

            X_ = compute_denoised_data( rank, X )
            S_  += X_
            S_2 += X_**2

            Y_ = np.dot(W, X)
            U_  += Y_
            U_2 += Y_**2

        S /= num_iter
        DE = np.sqrt( ( S2 - num_iter * S**2 )/( num_iter - 1 ) )

        S_ /= num_iter
        DE_ = np.sqrt( ( S_2 - num_iter * S_**2 )/( num_iter - 1 ) )

        U_ /= num_iter
        WE_ = np.sqrt( ( U_2 - num_iter * U_**2 )/( num_iter - 1 ) )

        return DE, DE_, WE_

    def generate_noise( self ):
        X = np.random.randn( *list(self.M.shape) )
        return X * self.E

def debug_plot_error( M, E_ ):
    import molass_legacy.KekLib.DebugPlot as plt
    x = np.arange(E_.shape[0])
    knots = np.linspace(x[0], x[-1], NUM_KNOTS)[1:-1]
    # for j in range(M.shape[1]):
    for j in [85, 210]:
        fig = plt.figure(figsize=(12,6))
        ax1 = fig.add_subplot(121)
        ax2 = fig.add_subplot(122)
        ax1.set_title('original curve and a sample simulated noisy curve in $Log_{10}$ scale')
        ax2.set_title('abs(noise) in $Log_{10}$ scale')
        y_ = M[:,j]
        e_ = E_[:,j]
        ax1.plot(np.log10(y_), label='data')
        ax1.plot(np.log10(y_ + e_), label='data + noise')
        ae = np.log10(np.abs(e_))
        spline = LSQUnivariateSpline( x, ae, knots )
        ax2.plot(ae, color='yellow', label='noise')
        ax2.plot(spline(x), ':', color='red', label='noise level')
        ax1.legend()
        ax2.legend()
        fig.tight_layout()
        try:
            plt.show()
        except:
            break

def theory_simulation_2D():
    noise_ratio = 0.05
    x = np.linspace(0.001, 1.0, 100)
    size = len(x)
    y = np.exp( -x*4 )
    e = np.random.randn(size)* y * noise_ratio
    def compute_simulated_curve(N):
        s_ = np.zeros(size)
        v_ = np.zeros(size)
        for i in range(N):
            e_ = np.random.randn(size)* y * noise_ratio
            y_ = y + e_
            s_ += y_
            v_ += y_**2
        Y = s_/N
        std = np.sqrt(v_/N - Y**2)
        return Y, std

    num_iter1 = 10
    Y1, std1 = compute_simulated_curve(num_iter1)
    num_iter2 = 100
    Y2, std2 = compute_simulated_curve(num_iter2)

    import molass_legacy.KekLib.DebugPlot as plt
    fig = plt.figure(figsize=(21,7))
    ax1 = fig.add_subplot(131)
    ax2 = fig.add_subplot(132)
    ax3 = fig.add_subplot(133)
    ax1.set_title('%d%% gaussian noise (one sample)' % (noise_ratio*100), fontsize=16)
    ax2.set_title('%d%% gaussian noise (result of %d random iterations)' % (noise_ratio*100, num_iter1),  fontsize=16)
    ax3.set_title('%d%% gaussian noise (result of %d random iterations)' % (noise_ratio*100, num_iter2),  fontsize=16)
    ax2t = ax2.twinx()
    ax3t = ax3.twinx()
    ax1.set_ylim(-0.1, 1.1)
    ax1.plot(x, y, label='exp(-x*4)')
    ax1.plot(x, e, label='5% gaussian noise')
    ax1.plot(x, y+e, label='exp(-x*4) + 5% gaussian noise')
    ax1.legend()
    ax2.plot(x, y, label='exp(-x*4)')
    ax2.plot(x, std1, label='STDEV(exp(-x*4))')
    ax2.plot(x, Y1, label='AVERAGE(exp(-x*4))')
    ax2t.set_ylim(0, 0.11)
    ax2t.plot(x, std1/Y1, ':', label='noise rate (STDEV/AVERAGE)')
    ax2.legend()
    ax2t.legend(bbox_to_anchor=(1, 0.85), loc='upper right')
    ax3.plot(x, y, label='exp(-x*4)')
    ax3.plot(x, std2, label='STDEV(exp(-x*4))')
    ax3.plot(x, Y2, label='AVERAGE(exp(-x*4))')
    ax3t.set_ylim(0, 0.11)
    ax3t.plot(x, std2/Y2, ':', label='noise rate (STDEV/AVERAGE)')
    ax3.legend()
    ax3t.legend(bbox_to_anchor=(1, 0.85), loc='upper right')
    fig.tight_layout()
    plt.show()

import molass_legacy.KekLib.DebugPlot as plt
import matplotlib.animation as animation

class McAnimation:
    def __init__(self):
        # import matplotlib.pyplot as plt
        self.pause = False

        noise_ratio = 0.05
        x = np.linspace(0.001, 1.0, 100)
        size = len(x)
        y = np.exp( -x*4 )

        num_iter = 1000
        e_list = []
        for i in range(num_iter):
            e = np.random.randn(size)* y * noise_ratio
            e_list.append(e)

        self.x = x
        self.y = y
        self.e_list = e_list
        self.num_iter = num_iter

        fig = plt.figure(figsize=(14,7))
        ax1 = fig.add_subplot(121)
        ax2 = fig.add_subplot(122)
        plt.dp.mpl_canvas.mpl_connect('button_press_event', self.on_click)

        ax1.set_title("Histogram of %d errors (noise ratio %.1g%%) at x=0" % (num_iter, noise_ratio*100), fontsize=16)
        ax2.set_title("Montecarlo simulation of noisy $e^{-4x}$  by %d iterations" % num_iter, fontsize=16)

        e_array = np.array(e_list)
        ax1.hist(e_array[:,0])
        ax2.plot(x, y, label='$e^{-4x}$')

        e = e_list[0]
        self.e_curve, = ax2.plot(x, e, label='%.1g%% noise' % (noise_ratio*100))
        self.y_curve, = ax2.plot(x, y+e, label='$e^{-4x}$ + noise')
        self.text = ax2.text(0.5,0.5, "", fontsize=80, alpha=0.3)

        ax2.legend(fontsize=16)
        fig.tight_layout()
        self.anim = animation.FuncAnimation(fig, self.animate, self.index_generator, init_func=self.anim_init,
                        interval=100, blit=True)
        # anim.save("Noise-MonteCarlo.mp4")
        self.fig = fig
        self.ax1 = ax1
        self.ax2 = ax2
        plt.show()

    def index_generator(self):
        i = 0
        while i < self.num_iter:
            i_ = i
            if self.pause:
                pass
            else:
                i += 1
            yield i_

    def on_click(self, event):
        if event.inaxes is None:
            return
        self.pause ^= True

    def anim_init(self):
        x = self.x
        y = self.y
        e = self.e_list[0]
        self.e_curve.set_xdata(np.ma.array(x, mask=True))
        self.e_curve.set_ydata(np.ma.array(e, mask=True))
        self.y_curve.set_xdata(np.ma.array(x, mask=True))
        self.y_curve.set_ydata(np.ma.array(y+e, mask=True))
        self.text.set_text("")
        return (self.e_curve, self.y_curve, self.text)

    def animate(self, i):
        x = self.x
        y = self.y
        e = self.e_list[i]
        self.e_curve.set_xdata(x)
        self.e_curve.set_ydata(e)
        self.y_curve.set_xdata(x)
        self.y_curve.set_ydata(y+e)
        self.text.set_text(str(i))
        return (self.e_curve, self.y_curve, self.text)

def simulated_svd_error_propagation_plot(x, M, E, M_, E_, SE, PE, WE_, eno, formula=True, mpmc=False):
    import molass_legacy.KekLib.DebugPlot    as plt

    if formula:
        fig = plt.figure(figsize=(21,7))
        ax0  = fig.add_subplot(131)
        ax1  = fig.add_subplot(132)
        ax2  = fig.add_subplot(133)
    else:
        fig = plt.figure(figsize=(14,7))
        ax1  = fig.add_subplot(121)
        ax2  = fig.add_subplot(122)

    if formula:
        ax0.set_title("Error propagation in denoising with SVD using MR-inverve")
        ax0.axes.get_xaxis().set_ticks([])
        ax0.axes.get_yaxis().set_ticks([])
        ax0.text(0.1,0.85, r"$M = U \cdot \Sigma \cdot V^T$", fontsize=30)
        ax0.text(0.1,0.65, r"$\widetilde{M} = \widetilde{U} \cdot \widetilde{\Sigma} \cdot  \widetilde{V}^T $", fontsize=30)
        ax0.text(0.1,0.45, r"$\widetilde{M} \approx  W \cdot M$", fontsize=30)
        ax0.text(0.1,0.25, r"$W \approx  W \cdot M \cdot M^+ \approx  \widetilde{M} \cdot M^+$", fontsize=30)
        ax0.text(0.1,0.05, r"$\widetilde{E} = \sqrt{ W^{(2)} \cdot E^{(2)} }$", fontsize=30)

    """
        [ a b ]  [x]
        [ c d ]  [y]
        ax + by
        cx + dy
    """

    ax1.set_title( "Propagated errors in linear scale at eno=%d" % eno )
    ax1.set_xlabel('Q')
    ax1.set_ylabel('Intensity')
    ax1.plot(x, M[eno, :], label='data')
    ax1.plot(x, E[eno, :], label='error')
    ax1.plot(x, M_[eno, :], label='denoised data')
    ax1.plot(x, E_[eno, :], label='error (MC)')
    ax1.plot(x, SE[eno, :], label='error of denoised data (MC)')
    ax1.plot(x, PE[eno, :], label='error of denoised data (matrix product)')
    if mpmc:
        ax1.plot(x, WE_[eno, :], label='error of denoised data (matrix product MC)')

    ax1.legend()
    ax2.set_title( "Propagated errors in $Log_{10}$ scale at eno=%d" % eno )
    ax2.set_xlabel('Q')
    ax2.set_ylabel('$Log_{10}(Intensity)$')
    ax2.plot(x, np.log10(M[eno, :]), label='data')
    ax2.plot(x, np.log10(E[eno, :]), label='error')
    ax2.plot(x, np.log10(M_[eno, :]), label='denoised data')
    ax2.plot(x, np.log10(E_[eno, :]), label='error (MC)')
    ax2.plot(x, np.log10(SE[eno, :]), label='error of denoised data (MC)')
    ax2.plot(x, np.log10(PE[eno, :]), label='error of denoised data (matrix produc)')
    if mpmc:
        ax2.plot(x, np.log10(WE_[eno, :]), label='error of denoised data (matrix produc MC)')
    ax2.legend()
    fig.tight_layout()
    plt.show()



"""
    DecompPlotUtils.py

    Copyright (c) 2023-2024, SAXS Team, KEK-PF
"""
import copy
import numpy as np
from molass_legacy.KekLib.OurMatplotlib import get_color
from DevSettings import get_dev_setting

def draw1_impl(self):
    ax  = self.ax1
    ax.cla()
    fx  = self.fx
    x   = self.x
    if self.elution_fig_type == 0:
        y   = self.y
        color = 'orange'
        opt_recs = self.opt_recs
    else:
        y   = self.uv_y
        color = 'blue'
        opt_recs = self.opt_recs_uv

    self.the_curves = []

    label = 'data'
    ax.plot(x, y, color=color, label=label)
    self.the_curves.append([label, x, y])

    total_y = np.zeros(len(y))
    resid_y = copy.deepcopy(y)

    cy_list = []
    areas = []
    for k, rec in enumerate(opt_recs):
        func = rec[1]
        y_ = func(fx)
        cy_list.append(y_)
        areas.append(np.sum(y_))
        self.the_curves.append([label, x, y_])
        total_y += y_
        resid_y -= y_

    props = np.array(areas)/np.sum(areas)
    for k, y_ in enumerate(cy_list):
        label = 'component-%d, (%.2f)' % ((k+1), props[k])
        ax.plot(x, y_, ':', label=label, color=get_color(k), linewidth=3)

    label = 'model total'
    ax.plot(x, total_y, ':', label=label, color='red', linewidth=3)
    self.the_curves.append([label, x, total_y])

    ax.legend()
    self.fig1_range_parts = []
    self.add_range_patchs(ax, self.fig1_range_parts)

    # draw the residual
    ax1r = self.ax1r
    ax1r.cla()
    label = 'residual'
    ax1r.plot(x, resid_y, label=label, color='gray')
    self.the_curves.append([label, x, resid_y])

    ymin, ymax = ax.get_ylim()
    yminr, ymaxr = ax1r.get_ylim()
    ymidr = (yminr + ymaxr)/2
    half_height_r = max((ymax - ymin)*self.height_ratio, ymaxr - yminr)/2
    ax1r.set_ylim(ymidr - half_height_r, ymidr + half_height_r)

    ax1r.legend(bbox_to_anchor=(1, 1), loc='lower right')
    self.fig1_range_parts_resid = []
    self.add_range_patchs(ax1r, self.fig1_range_parts_resid)

    self.fit_error = np.sum(np.abs(resid_y))

    if get_dev_setting('running_with_tester'):
        self.write_residual_amount_to_testerlog(y, resid_y)



import json
import os

path = os.path.dirname(__file__)

with open(os.path.join(path, "electrons.json"), 'r') as f:
    electrons = json.load(f)

with open(os.path.join(path, "mean_volumes.json"), 'r') as f:
    atomic_volumes = json.load(f)

with open(os.path.join(path, "numH.json"), 'r') as f:
    numH = json.load(f)

with open(os.path.join(path, "volH.json"), 'r') as f:
    volH = json.load(f)

with open(os.path.join(path, "vdW.json"), 'r') as f:
    vdW = json.load(f)

with open(os.path.join(path, "radii_scale_factors.json"), 'r') as f:
    radii_sf_dict = json.load(f)

with open(os.path.join(path, "cromer_mann_4gauss.json"), 'r') as f:
    ffcoeff = json.load(f)




"""
    Absorbance.py

    Copyright (c) 2017-2023, SAXS Tam, KEK-PF
"""
import numpy                as np
import copy
from scipy                  import stats
from bisect                 import bisect_right
import logging
from molass_legacy.KekLib.ChangeableLogger import arg_join
# from lmfit                  import minimize, Parameters
from molass_legacy.KekLib.LmfitThreadSafe import minimize, Parameters
from molass_legacy._MOLASS.SerialSettings import get_setting, INTEGRAL_BASELINE
from molass_legacy.SerialAnalyzer.ElutionCurve           import ElutionCurve
from molass_legacy.Baseline.ScatteringBaseline import ScatteringBaseline
from molass_legacy.SerialAnalyzer.ScatteringBasecurve import ScatteringBasecurve
from molass_legacy.SerialAnalyzer.ScatteringBasespline import ScatteringBasespline
from molass_legacy.SerialAnalyzer.BasePlane import LowPercentilePlane, LambertBeerPlane
from molass_legacy.SerialAnalyzer.UvBaseSurface import UvBaseSurface
from molass_legacy.Test.TesterLogger import write_to_tester_log
from molass_legacy.KekLib.ThreeDimUtils import compute_plane
import molass_legacy.KekLib.DebugPlot as plt

POSITVE_SMALL_VALUE     = 1e-5
MIN_ASC_WIDTH           = 10
USABLE_JUMP_POINT_RATIO = 0.1
LOW_PERCENTILE_OK_RATIO = 0.8       # not ok for HIF
LOW_PERCENTILE_BASE_1ST = 30
LOW_PERCENTILE_BASE_FIN = 10
USE_SUB_PICKING_IN_LPM  = True
END_ABNORMALITY_RATIO   = 5
ALLOWED_SHIFTED_RATIO   = 0.1
TAIL_BAND_WIDTH = 11

def compute_scatter( S, x, z ):
    denom = np.max( np.vstack( [ np.ones( len(x) )*1e-10, 1 - S * np.power( x, -4 ) ] ), axis=0 )
    return np.dot( np.log( 1 / denom ).reshape( ( len(x), 1 ) ), z.reshape( ( 1, len(z) ) ) )

def non_zero_value( v ):
    if v >= 0:
        v_ = max( POSITVE_SMALL_VALUE, v )
    else:
        v_ = min( -POSITVE_SMALL_VALUE, v )
    return v_

class FixedBaseResult:
    def __init__( self, fidex_base, fixed_slope, size, j ):

        if fixed_slope is None:
            A   = 0
            B   = 0
            C   = fidex_base
        else:
            A   = 0
            B   = fixed_slope
            j_  = 0  if j is None else j
            C   = fidex_base - fixed_slope * j_

        self.params  = params = Parameters()
        params.add( 'A', value=A, vary=False )
        params.add( 'B', value=B, vary=False )
        params.add( 'C', value=C, vary=False )
        params.add( 'S', value=0, vary=False )

class Absorbance:
    def __init__( self, wl_vector, data, x_curve,
            col_header=None,
            std_wvlen=None, end_wvlen=None,
            scattering_base=None, independent=False, use_lpm=False,
            orig_top_x=None,
            debug=False ):
        self.debug      = debug
        self.logger     = logging.getLogger( __name__ )
        wl_slice_ = self.check_abnormality_in_wvlen_ends(data, wl_vector)
        self.wl_vector  = wl_vector[wl_slice_]
        self.data       = data[wl_slice_, :]
        """
        Due to thie slicing,
        be aware that self.data is not always equals to the input data,
        i.e., serial_data.conc_array.
        """
        self.x_curve    = x_curve
        self.orig_top_x = orig_top_x
        self.col_header = col_header
        self.better_mapping_info = None
        if scattering_base is None:
            scattering_base = get_setting( 'scattering_base' )
        # self.independent = ( independent or scattering_base == 0 )
        self.independent = independent
        self.rail_points    = None

        self.jvector    = np.arange( self.data.shape[1] )
        self.update_optimizer_range( std_wvlen, end_wvlen )
        self.mapped_intensity = None
        # self.compute_base_curve( use_lpm )
        # write_to_tester_log( 'real_flow_changes=' + str( self.get_real_flow_changes() ) + '\n' )

        # print( 'Absorbance: jump_j=', self.jump_j )

    def update_optimizer_range( self, std_wvlen, end_wvlen ):
        from molass_legacy.UV.PlainCurveUtils import get_flat_wavelength

        self.std_wvlen  = get_setting( 'absorbance_picking' ) if std_wvlen is None else std_wvlen
        self.end_wvlen  = get_flat_wavelength(self.wl_vector) if end_wvlen is None else end_wvlen
        self.reg_wvlens = np.linspace( self.std_wvlen, self.end_wvlen, 4 )
        # print( 'reg_wvlens=', self.reg_wvlens )

        a_vector, i     = self.get_vector_at( self.std_wvlen )
        self.index      = i
        self.a_vector   = a_vector
        self.istd       = bisect_right( self.wl_vector, self.std_wvlen )
        if USE_SUB_PICKING_IN_LPM:
            self.istd_sub   = bisect_right( self.wl_vector, get_setting('absorbance_picking_sub') )
        self.iend       = bisect_right( self.wl_vector, self.end_wvlen )

        # take min to copw with cases where wl_vector has been trimmed
        self.i400       = min(len(self.wl_vector) - TAIL_BAND_WIDTH, bisect_right(self.wl_vector, 400))

        self.a_curve    =  ElutionCurve( self.a_vector, orig_top_x=self.orig_top_x )

        wvlen_sub = get_setting( 'absorbance_picking_sub' )
        a_vector_sub, i_sub = self.get_vector_at( wvlen_sub )
        self.index_sub = i_sub
        self.a_vector_sub = a_vector_sub
        self.a_curve_sub =  ElutionCurve( self.a_vector_sub, possiblly_peakless=True )

    def check_abnormality_in_wvlen_ends( self, data, wl_vector):
        num_points = 5
        start_sumv_array = np.array([ np.sum(np.abs(data[i,:])) for i in range(num_points) ])
        average_len = 3

        self.safe_start = None
        last_average = None
        for k in range(num_points - average_len + 1):
            average = np.average(start_sumv_array[k:k+average_len])
            if last_average is not None:
                ratio = last_average / average
                # print([k], ratio)
                if ratio > END_ABNORMALITY_RATIO:
                    self.safe_start = k

            last_average = average
        if self.safe_start is not None:
            i = self.safe_start - 1
            self.logger.warning("abnormal values detected at wave length[%d]=%g" % (i, wl_vector[i]))

        size = data.shape[0]
        final_sumv_array = np.array([ np.sum(np.abs(data[i,:])) for i in range(size-num_points, size) ])
        self.safe_end = None
        last_average = None
        for k in reversed(range(num_points - average_len + 1)):
            average = np.average(final_sumv_array[k:k+average_len])
            if last_average is not None:
                ratio = last_average / average
                # print([k], ratio)
                if ratio > END_ABNORMALITY_RATIO:
                    self.safe_end = size - num_points + k
            last_average = average

        if self.safe_end is not None:
            i = self.safe_end + 1
            self.logger.warning("abnormal values detected at wave length[%d]=%g" % (i, wl_vector[i]))

        return slice(self.safe_start, self.safe_end)

    def get_wave_len_ends( self ):
        return self.safe_start, self.safe_end

    def get_shifted_elution_base( self ):
        # compute_base_curve must have been called for self.rail_points
        fc = self.rail_points
        m_base  = np.zeros( fc[1]-fc[0]+1 ) if self.base_curve is None else self.base_curve
        l_base  = np.ones( fc[0] ) * m_base[0]
        r_base  = np.ones( len(self.a_vector) - fc[1] -1 ) * m_base[-1]
        base    = np.hstack( [ l_base, m_base, r_base ] )

        if False:
            print("len(base), len(self.a_vector)", len(base), len(self.a_vector))
            plt.push()
            fig, ax = plt.subplots()
            ax.plot(self.a_vector)
            ax.plot(base)
            plt.show()
            plt.pop()

        assert len(base) == len(self.a_vector)
        return base

    def get_standard_elution_base( self ):
        return self.get_bottomline_vector()

    def compute_base_curve( self, pre_recog, baseline_type, full_width=True, debug=False ):
        """
        full_width=True because this baseline_type should be called after trimming
        """

        if full_width:
            # temp fix to avoid getting false j_min, j_max
            j_min, j_max = 0, len(self.a_vector)-1
        else:
            fc = pre_recog.flowchange
            j_min, j_max1 = fc.get_flow_changes()
            # j_max1 can be the length
            j_max = j_max1 - 1

        self.logger.info('computing base curve (type %d)  for UV data with j_min=%d, j_max=%d in data.shape=%s' % (baseline_type, j_min, j_max, str(self.data.shape)))

        base_curve_done = False

        if baseline_type == 0:
            self.base_curve = np.zeros(j_max + 1 - j_min)
            self.baseplane_params = None
            base_curve_done = True
        elif baseline_type == 1:
            self.solve_bottomplane_LPM(j_min, j_max, debug=debug)
        elif baseline_type == 4:
            self.solve_bottomplane_LB(j_min, j_max, debug=debug)
        elif baseline_type == 5:
            self.compute_integral_basecurve()
            base_curve_done = True
        else:
            assert False

        if base_curve_done:
            return

        self.rail_points    = [ j_min, j_max ]

        # print( 'compute_base_curve: len(self.a_curve.peak_info)=', len(self.a_curve.peak_info) )

        if False:
            self.rail_points_sigma  = [ self.a_curve.compute_end_coordinate_in_sigma(i) for i in self.rail_points ]
            print( 'rail_points_sigma=', self.rail_points_sigma )
        else:
            self.rail_points_sigma  = None

        A, B, C = self.get_baseplane_params()
        estimeted_rails_base    = [ A*self.std_wvlen + B*i + C  for i in self.rail_points ]
        """
        self.rail_points_offset = [ ( self.a_curve.y[i] - ( A*self.std_wvlen + B*i + C ) ) / self.a_curve.max_y for i in self.rail_points ]
        print( 'rail_points_offset=', self.rail_points_offset )
        """

        self.tail_band      = slice( self.i400, self.i400+TAIL_BAND_WIDTH )
        self.tail_index     = self.i400 + 5
        self.tail_curve     = np.average( self.data[ self.tail_band, j_min:j_max+1 ], axis=0 )
        rails = [
                self.data[ [ self.index, self.tail_index ], j_min ],
                self.data[ [ self.index, self.tail_index ], j_max ],
                ]
        try:
            base_surface  = UvBaseSurface( self.rail_points, self.tail_curve, rails,
                                        self.rail_points_sigma,
                                        self.a_curve.max_y,
                                        estimeted_rails_base,
                                        )
            self.base_curve = base_surface.get_basecurve()
        except RuntimeError:
            self.base_curve = None

        return self.base_curve

    def get_integral_basecurve(self):
        return self.base_curve

    def get_vector_at( self, wvlen ):
        i = min(len(self.wl_vector) - 1, bisect_right( self.wl_vector, wvlen ))
        return np.array( self.data[i, :] ), i

    def get_standard_vector( self ):
        return self.a_vector

    def get_corrected_vector( self, absorbance_baseline_type, scattering_base, baseline_degree, params=None ):
        slice_size = self.right_jump_j_safe - self.jump_j_safe + 1

        # scattering_base = get_setting( 'scattering_base' )
        # baseline_degree = get_setting( 'baseline_degree' )
        # absorbance_baseline_type = get_setting( 'absorbance_baseline_type' )

        if absorbance_baseline_type == 0 or self.base_curve is None:
            if scattering_base == 2 and slice_size / len(self.a_vector) > LOW_PERCENTILE_OK_RATIO:
                a_  = self.a_vector[self.jump_j_safe:self.right_jump_j_safe+1]
                j   = self.jvector - self.jump_j_safe
                if baseline_degree == 1:
                    abl = ScatteringBaseline( a_ )
                    A, B = abl.solve()
                    self.bottomline = A*j + B
                elif baseline_degree == 2:
                    abl = ScatteringBasecurve( a_ )
                    A, B, C = abl.solve()
                    self.bottomline = A*j**2 + B*j + C
                else:
                    abl = ScatteringBasespline( a_ )
                    abl.solve()
                    self.bottomline = abl.get_baseline( j )
            else:
                self.bottomline = self.get_bottomline_vector( params=params )
        else:
            self.bottomline = np.hstack( [
                            np.ones( self.jump_j_safe ) * self.base_curve[0],
                            self.base_curve,
                            np.ones( len(self.a_curve.y) - 1 - self.right_jump_j_safe ) * self.base_curve[-1],
                            ] )

        y = self.a_vector - self.bottomline
        y[ y < 0 ] = 0
        y[ 0:self.jump_j_safe ] = 0
        return y

    def compute_scatter_factor_at( self, A, B, C, i ):
        peak_j  = self.a_curve.middle_abs
        end_wvlen_ = self.wl_vector[i]
        std_base = A * self.std_wvlen + B * peak_j + C
        end_base = A * end_wvlen_ + B * peak_j + C

        """
            Abs_ = log( 1 / ( 1 - S*λ⁻⁴ ) ) * Abs0
            exp( Abs_ / Abs0 ) = 1 / ( 1 - S*λ⁻⁴ )
            1 - S*λ⁻⁴ = 1 / exp( Abs_ / Abs0 ) = exp( - Abs_ / Abs0 )
            1 - exp( - Abs_ / Abs0 ) = S*λ⁻⁴
            S = (  1 - exp( - Abs_ / Abs0 ) ) * λ⁴
        """
        S = ( 1 - np.exp( - (self.data[ i, peak_j ] - end_base )
                                    / ( self.a_vector[peak_j] - std_base )
                              )
            ) * np.power( end_wvlen_, 4 )
        return S

    def get_jump_point_base( self ):
        A, B, C = self.get_baseplane_params()
        z = A*self.std_wvlen + B*self.jump_j_safe + C
        return z

    def get_mapped_slope_and_intercept( self ):
        A, B, C = self.get_baseplane_params()
        slope     = B
        intercept = A * self.std_wvlen + C
        return slope, intercept

    def get_conc_params( self, conc_factor ):
        A, B, C = self.get_baseplane_params()
        conc_slope     = conc_factor * B
        conc_intercept = conc_factor * ( A * self.std_wvlen + C )
        print( __name__ + '.get_conc_params: conc_slope=', conc_slope, 'conc_intercept=', conc_intercept )
        return conc_slope, conc_intercept

    def get_bottomline_vector( self, params=None ):
        if params is None:
            A, B, C = self.get_baseplane_params()
        else:
            A, B, C = params['A'], params['B'], params['C']
        # print( 'get_bottomline_vector: S=', S.value )
        jv  = self.jvector
        base =  A * self.std_wvlen + B * jv + C
        if False:
            print( 'get_bottomline_vector(%d): A=%.4g, B=%.4g, C=%.4g' % (id(self), A, B, C ), params)
            fig = plt.figure()
            ax = fig.gca()
            ax.set_title("get_bottomline_vector: debug")
            ax.plot(jv, self.a_vector, color='blue')
            ax.plot(jv, base, color='red')
            plt.show()
        return base

    def get_bottomline_vector_near_peak( self ):
        A, B, C = self.get_baseplane_params()
        lower   = self.a_curve.lower_abs
        middle  = self.a_curve.middle_abs
        upper   = self.a_curve.upper_abs
        jv  = np.array( [ lower, middle, upper ] )
        ret_vectors = compute_plane( A, B, C, self.reg_wvlens, jv )
        z0 = self.a_vector[jv] - ret_vectors[0,:]
        ret_scatter = compute_scatter( S, self.reg_wvlens, z0 )

        return ( ret_vectors + ret_scatter ).T

    def get_wv_index_vector( self ):
        iv = []
        for wvlen in self.reg_wvlens:
            iv.append( bisect_right( self.wl_vector, wvlen ) )
        self.index_vector = np.array( iv )
        return self.index_vector

    def get_bottomline_matrix( self ):
        A, B, C = self.get_baseplane_params()
        jv  = self.jvector
        ret_matrix  = compute_plane( A, B, C, self.reg_wvlens, jv )
        z0 = self.a_vector - ret_matrix[0,:]
        ret_scatter = compute_scatter( 0, self.reg_wvlens, z0 )
        return ( ret_matrix + ret_scatter).T

    def get_normal_region( self ):
        return self.jump_j_safe, self.right_jump_j_safe

    def solve_bottomplane_LPM( self, j_min, j_max, debug=False ):
        iy  = slice( j_min, j_max+1 )
        uv_lpm_option = get_setting('uv_lpm_option')
        # suppress 3D-LPM temporarily
        if uv_lpm_option == 1 and False:
            if USE_SUB_PICKING_IN_LPM:
                start = self.istd_sub
            else:
                start = self.istd
            ix  = slice( start, self.iend )
            lpp = LowPercentilePlane(self.data, self.wl_vector, self.a_vector, ix, iy, self.index, debug)
            params = lpp.get_params()
        else:
            from molass_legacy.SerialAnalyzer.Uv2dLpm import Uv2dLpm
            lpm = Uv2dLpm(self.a_vector, iy)
            params = lpm.get_params()

        option_text = '3D' if uv_lpm_option == 1 else '2D'
        self.logger.info('UV baseplane_params are determined to %s with %s option.', str(params), option_text)
        self.baseplane_params = params

    def get_baseplane_params( self ):
        return self.baseplane_params

    def solve_bottomplane_LB( self, j_min, j_max,
            init_values=None, fixed_base=None, fixed_slope=None, left_end_j=None,
            debug=False ):

        if USE_SUB_PICKING_IN_LPM:
            start = self.istd_sub
        else:
            start = self.istd
        ix  = slice( start, self.iend )
        iy  = slice( j_min, j_max+1 )
        istd_ = self.istd - ix.start

        lbp = LambertBeerPlane(self.data, self.wl_vector, ix, iy, self.a_vector, self.index, istd_, debug)
        self.baseplane_params   = lbp.get_params()
        print('solve_bottomplane_LB(%d)' % id(self), self.baseplane_params)

    def get_corrected_data( self ):
        A, B, C = self.baseplane_params
        x   = self.wl_vector 
        y   = self.jvector
        base_plane = compute_plane( A, B, C, x, y )
        data_ = self.data - base_plane
        return data_

    def get_zlim_for_3d_plot( self ):
        start = bisect_right( self.wl_vector, 210 )
        ix  = slice( start, self.iend+1 )
        iy  = slice( self.jump_j_safe, self.right_jump_j_safe+1 )
        data_ = self.data[ix, iy]
        zmin = np.min(data_)
        zmax = np.max(data_)
        return zmin, zmax

    def compute_integral_basecurve(self):
        from molass_legacy.Baseline.Baseline import better_integrative_curve
        end_slices = self.a_curve.get_end_slices()
        self.base_curve, convex = better_integrative_curve(self.a_vector, end_slices=end_slices)
        self.logger.info("computed UV basecurve with better_integrative_curve%s.", " using convex method" if convex else "")
        return self.base_curve

    def shifted_baseline_ok(self):
        lower, center, upper = self.a_curve.get_primary_peak_info()
        peak_top_ratio = self.a_curve_sub.y[center]/self.a_curve.y[center]
        ret = peak_top_ratio < ALLOWED_SHIFTED_RATIO
        if not ret:
            self.logger.info("peak_top_ratio=%.3g at λ=%g over λ=%g.", peak_top_ratio, self.end_wvlen, self.std_wvlen)
        return ret



# -*- coding: utf-8 -*-
"""

    ファイル名：   Preferences.py

    処理内容：

        オプション情報

"""
from __future__ import division, print_function, unicode_literals

from PersistentInfo         import PersistentInfo

DEFAULT_PREFERENCES = {
    'syn_method'        : 'cover',
    'detection_counter' : 'None',
    'postfix_syn'       : '_syn',
    'color_map'         : 'ALBULA',
    'save_policy'       : 'Ask',
    'syn_policy'        : 'all',
    'syn_flags'         : [ 1, 1, 1 ],
    }

preference_file = 'preferences.dump'
preference_info = PersistentInfo( preference_file, DEFAULT_PREFERENCES )
preferences_    = preference_info.get_dictionary()

def reload_preferences():
    global preference_info
    global preferences_

    preference_info    = PersistentInfo( preference_file, DEFAULT_PREFERENCES )
    preferences_       = preference_info.get_dictionary()

def clear_preferences():
    global preferences_
    preferences_ = {}
    preference_info.set_dictionary( preferences_ )

def get_preference( item ):
    assert( item in DEFAULT_PREFERENCES )
    return preferences_.get( item )

def set_preference( item, value ):
    assert( item in DEFAULT_PREFERENCES )
    preferences_[ item ] = value

def temporary_preferences_begin():
    global preferences_save
    preferences_save = preferences_.copy()

def temporary_preferences_end():
    global preferences_save
    global preferences_
    preferences_ = preferences_save
    preference_info.set_dictionary( preferences_ )

def get_usual_preference( item ):
    global preferences_save
    return preferences_save.get( item )

def save_preferences():
    preference_info.save()



import pymc as pm

def reproduce_bug():
    print(pm.__version__)

    def _logp(value, mu):
        return -((value - mu) ** 2)

    def _random(mu, rng=None, size=None):
        return rng.normal(loc=mu, scale=1, size=size)

    with pm.Model():
        mu = pm.CustomDist("mu", 0, logp=_logp, random=_random)
        pm.CustomDist("y", mu, logp=_logp, class_name="", random=_random, observed=[1, 2])
        pm.sample_smc(draws=6, cores=2)

if __name__ == '__main__':
    reproduce_bug()


# coding: utf-8
"""
    AbsorbancePlot.py

    Copyright (c) 2017-2018, Masatsuyo Takahashi, KEK-PF
"""
import numpy                as np
from bisect                 import bisect_right
import matplotlib.pyplot    as plt
import matplotlib.gridspec  as gridspec
from mpl_toolkits.mplot3d   import Axes3D
from molass_legacy.UV.Absorbance             import Absorbance
from molass_legacy.SerialAnalyzer.AnalyzerUtil           import compute_conc_factor_util
from molass_legacy._MOLASS.SerialSettings         import get_setting

DEBUG = False

def get_temp_absorbance( dialog, serial_data ):
    serial_data.wait_until_ready()

    abs_ = serial_data.absorbance
    std_wvlen = dialog.absorbance_picking.get()
    try:
        end_wvlen = dialog.zero_absorbance.get()
    except:
        end_wvlen = None

    temp_absorbance = Absorbance( abs_.wl_vector, abs_.data, abs_.i_curve, std_wvlen=std_wvlen, end_wvlen=end_wvlen )
    return temp_absorbance

def show_absorbance_figure_util( dialog, make_temp=False ):
    print( 'show_absorbance_figure_util' )
    dialog.config( cursor='wait' )
    dialog.update()

    if make_temp:
        dialog.temp_absorbance = absorbance = get_temp_absorbance( dialog, dialog.parent.serial_data )
    else:
        absorbance = dialog.parent.serial_data.absorbance

    conc_factor = compute_conc_factor_util()
    # absorbance.apply_mapping( conc_factor, dialog.parent.serial_data.xray_curve )
    # why this call ?

    plot = AbsorbancePlot( absorbance )
    plot.draw_3d_for_setting_gui( dialog, 'Absorbance from ' + dialog.parent.uv_folder.get() )
    dialog.config( cursor='' )

class AbsorbancePlot:
    def __init__( self, absorbance ):
        self.absorbance     = absorbance
        self.wl_vector      = absorbance.wl_vector
        self.wvlen_lower    = 245
        self.wvlen_upper    = 450
        f = bisect_right( absorbance.wl_vector, self.wvlen_lower )
        t = bisect_right( absorbance.wl_vector, self.wvlen_upper )
        self.i_slice        = slice( f, t )

    def draw_3d_absorbance( self, ax, data, title, absorbance_baseline_type=None, zlim=None ):

        lower   = self.wvlen_lower
        upper   = self.wvlen_upper

        ax.set_title( title )
        ax.set_xlabel( '\nwave length (nm)' )
        ax.set_ylabel( '\nsequential №' )
        ax.set_zlabel( '\nabsorbance' )
        ax.set_xlim( self.wvlen_lower - 10 , self.wvlen_upper + 10 )

        if zlim is None:
            # zlim = ( -0.05, 0.2 )
            # ax.set_zlim( zlim )
            pass
        else:
            ax.set_zlim( zlim )

        print( 'data.shape=', data.shape )

        size = data.shape[1]

        peak_j = None
        tail_i = self.absorbance.tail_index
        if absorbance_baseline_type is None:
            absorbance_baseline_type = get_setting( 'absorbance_baseline_type' )
        for i, wv in enumerate( self.absorbance.wl_vector ):
            if i == tail_i:
                pass
            else:
                if wv < self.wvlen_lower or wv > self.wvlen_upper:
                    continue
                if i != self.absorbance.index and i % 2 > 0:
                    continue

            X = np.ones( size ) * wv
            Y = np.arange( size )
            Z = data[i,:]
            if i == self.absorbance.index:
                alpha   = 1
                color   = 'blue'
            else:
                alpha   = 0.2
                color   = '#1f77b4'
            ax.plot( X, Y, Z, color=color, alpha=alpha )

            if absorbance_baseline_type == 0 or self.absorbance.base_curve is None:
                continue

            if i  == tail_i:
                slice_ = slice( self.absorbance.rail_points[0], self.absorbance.rail_points[1]+1 )
                Y_ = Y[slice_]
                curve_size = len(self.absorbance.tail_curve)
                for ii in range( i-5, i+6, 2 ):
                    X_ = np.ones( curve_size ) * self.absorbance.wl_vector[ii]
                    ax.plot( X_, Y_, self.absorbance.tail_curve, color='red', alpha=0.5 )

                X_ = np.ones( curve_size ) * self.absorbance.std_wvlen
                ax.plot( X_, Y_, self.absorbance.base_curve, color='red', alpha=0.5 )

    def draw_3d_absorbance_detail( self, ax, ax_2d ):
        wvlen   = self.absorbance.wl_vector
        lower   = self.wvlen_lower
        upper   = self.wvlen_upper
        absorbance  = self.absorbance

        wvlen_ = wvlen[self.i_slice]

        if False:
            lower_j  = absorbance.a_curve.lower_abs
            middle_j = absorbance.a_curve.middle_abs
            upper_j  = absorbance.a_curve.upper_abs
        else:
            n       = absorbance.a_curve.primary_peak_no
            info    = absorbance.a_curve.peak_info[n]
            lower_j  = info[0]
            middle_j = int(info[1] + 0.5)
            upper_j  = info[2]

        section_array = []
        colors = [ 'green', 'orange', 'green', 'purple', 'purple' ]
        for i, j in enumerate( [ lower_j, middle_j, upper_j, absorbance.jump_j, absorbance.right_jump_j ] ):
            print( 'draw_3d_absorbance_detail: (i, j)=', (i, j) )
            if j is None:
                continue

            X = wvlen_
            Y = np.ones( len(wvlen_) ) * j
            Z = self.absorbance.data[self.i_slice,j]
            section_array.append( [ j, Z ] )
            color = colors[i]
            ax.plot( X, Y, Z, ':', color=color )

        # self.draw_wave_length_side_2d( ax_2d,  wvlen_, section_array )

        bline = absorbance.get_bottomline_matrix()

        xmin, xmax = ax.get_xlim()
        zmin, zmax = ax.get_zlim()
        xoffset = ( xmax - xmin ) * 0.01
        zoffset = ( zmax - zmin ) * 0.005

        Y = np.arange( absorbance.data.shape[1] )
        for i in range(4):
            wvlen = absorbance.reg_wvlens[i]
            X = np.ones( absorbance.data.shape[1] ) * wvlen
            Z = bline[:, i]
            if i == 3:
                color   = 'cyan'
                alpha   = 1
            else:
                color   = 'red'
                alpha   = 0.5
            ax.plot( X, Y, Z, ':', color=color, alpha=alpha )
            if i in [0,3]:
                ax.text( wvlen - xoffset, Y[-1], Z[-1] + zoffset, 'λ=%g' % wvlen )

    def get_component( self, U, s, V, i ):
        # Be aware that we are slicing on the column of V, because V is (Vtt=V), not Vt.
        return s[ i ] * np.dot( np.transpose( [ U[ :, i ] ] ), np.array( [ V[ :, i ] ] ) )

    def draw_wave_length_side_2d( self, ax, wvlen, section_array ):
        ax.set_title( 'Absorbance at sequential sections' )
        ax.set_xlabel( 'Wave Length' )
        ax.set_ylabel( 'Absorbance' )
        ax.set_xlim( self.wvlen_lower - 10 , self.wvlen_upper + 10 )

        colors = [ 'green', 'orange', 'green', 'purple', 'purple' ]
        max_z = []
        labels = [ 'left', 'peak', 'right', 'jump point', 'jump point' ]

        m = np.argmax( section_array[1][1] )

        for i, rec in enumerate( section_array ):
            j, Z = rec
            line, = ax.plot( wvlen, Z, ':', color=colors[i],  label='seqno %d ( %s )' % ( j, labels[i] ) )
            max_z.append( Z[m] )

        A, B, C, S  = self.absorbance.get_params()
        b_points    = self.absorbance.get_bottomline_vector_near_peak()
        print( 'b_points.shape=', b_points.shape )
        j_peak, Z_peak = section_array[1]
        i0 = self.absorbance.index_vector[0] - self.i_slice.start
        peak_conc   = Z_peak[i0] - b_points[1][0]

        print( 'A, B, C=', A.value, B.value, C.value )
        print( 'b_points[1][0]=', b_points[1][0] )

        section_bottom = A * wvlen + B * j_peak + C
        ax.plot( wvlen, section_bottom, ':', color='pink' )

        for i in [ 0, 2 ]:
            if i >= len(max_z):
                continue

            j, Z = section_array[i]
            print( i, 'j=', j )
            section_bottom = A * wvlen + B * j + C
            Z_ = Z - section_bottom
            scale = peak_conc / Z_[i0]
            z = Z_ * scale + section_bottom
            ax.plot( wvlen, z, ':', color='yellow', label='seqno %d scaled' % j )
            ax.plot( wvlen, section_bottom, ':', color='pink' )

        x0  = self.absorbance.reg_wvlens[0]
        z0  = self.absorbance.get_standard_vector()
        y0_min = np.min( z0 )
        y0_max = np.max( z0 )
        ax.plot( [ x0, x0 ], [ y0_min, y0_max ], color='blue', alpha=0.5 )

        x = self.absorbance.reg_wvlens
        for i, bv in enumerate( b_points ):
            print( i, 'bv=', bv )
            ax.scatter( x[:-1], bv[:-1], color='red', alpha=0.5, s=20 )
            ax.scatter( x[-1], bv[-1], color='cyan', alpha=1, s=20 )

        ymin, ymax = ax.get_ylim()
        yoffset = ( ymax - ymin )*0.05
        y_ = np.min( b_points ) - yoffset*0.3

        for i in [0,3]:
            x_ = x[i]
            ax.annotate( 'λ=%g' % (x_), xy=(x_, y_),
                            xytext=( x_, y_ - yoffset ),
                            ha='center', va='center',
                            arrowprops=dict( headwidth=5, width=0.5, color='black', shrink=0.05),
                            )
        ax.legend()

    def draw_3d_with_svd( self, title ):

        U, s, Vt = np.linalg.svd( self.absorbance.data )
        V = np.transpose( Vt )

        components = []
        for i in range(3):
            components.append( self.get_component( U, s, V, i ) )

        fig = plt.figure( figsize=( 15, 9 ) )
        gs = gridspec.GridSpec( 3, 3 )
        ax1 = fig.add_subplot( gs[:, 0:2], projection='3d' )
        ax2 = fig.add_subplot( gs[0, 2], projection='3d' )
        ax3 = fig.add_subplot( gs[1, 2], projection='3d' )
        ax4 = fig.add_subplot( gs[2, 2], projection='3d' )

        self.draw_3d_absorbance( ax1, self.absorbance.data, title )

        fig_2d = plt.figure( figsize=( 8, 6 ) )
        ax_2d  = fig_2d.add_subplot( 111 )
        self.draw_3d_absorbance_detail( ax1, ax_2d )
        fig_2d.tight_layout()

        zlim = ax1.get_zlim()
        self.draw_3d_absorbance( ax2, components[0], "Component 0", zlim=zlim )
        self.draw_3d_absorbance( ax3, components[1], "Component 1", zlim=zlim )
        self.draw_3d_absorbance( ax4, components[2], "Component 2", zlim=zlim )

        fig.tight_layout()
        plt.show()

    def draw_3d_for_setting_gui( self, parent, title, absorbance_baseline_type=None ):
        from DebugCanvas            import DebugCanvas
        from molass_legacy.KekLib.TkUtils                import is_low_resolution

        def draw_func( fig ):
            # fig.set_size_inches( 20, 10 )
            # gs  = gridspec.GridSpec( 1, 2 )
            # ax1 = fig.add_subplot( gs[0,0], projection='3d' )
            # ax2 = fig.add_subplot( gs[0,1] )
            ax1 = fig.add_subplot( 111, projection='3d' )
            ax2 = None
            self.draw_3d_absorbance( ax1, self.absorbance.data, title, absorbance_baseline_type=absorbance_baseline_type )
            self.draw_3d_absorbance_detail( ax1, ax2 )
            fig.tight_layout()

        figsize = ( 9, 8 ) if is_low_resolution() else ( 12, 11 )
        self.canvas = DebugCanvas( "Absorbance Data 3D Plot", draw_func, parent=parent, figsize=figsize )
        self.canvas.show( cursor_update=False )



# coding: utf-8
"""
    SimultaneousLRF.py

    Copyright (c) 2021, SAXS Team, KEK-PF
"""
import logging
import numpy as np
from matplotlib.gridspec import GridSpec
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.Trimming.PreliminaryRecognition import PreliminaryRecognition

def concatinate_data(pre_recog, X, U):
    M = np.zeros((X.shape[0]+U.shape[0], X.shape[1]))
    M[0:X.shape[0],:] = X

    slope = pre_recog.cs.slope
    intercept = pre_recog.cs.intercept
    for j in range(X.shape[1]):
        f = max(0, int(round(slope*(j-0.5) + intercept)))
        t = min(U.shape[1], int(round(slope*(j+0.5) + intercept)))
        v = np.average(U[:,f:t], axis=1)
        M[X.shape[0]:,j] = v
    return M

def demo(sd):
    from MatrixData import simple_plot_3d
    U = sd.conc_array
    X, E, qv, ecurve = sd.get_xr_data_separate_ly()
    pre_recog = PreliminaryRecognition(sd)
    M = concatinate_data(pre_recog, X, U)

    plt.push()
    fig = plt.figure(figsize=(21,11))
    gs = GridSpec(4,4)
    ax1 = fig.add_subplot(gs[0:2,0], projection="3d")
    ax2 = fig.add_subplot(gs[2:4,0], projection="3d")
    ax3 = fig.add_subplot(gs[1:3,1], projection="3d")

    simple_plot_3d(ax1, X)
    simple_plot_3d(ax2, U)
    simple_plot_3d(ax3, M)

    fig.tight_layout()
    plt.show()
    plt.pop()



"""
    Processes.RgComputer.py

    Copyright (c) 2021, SAXS Team, KEK-PF
"""
from time import time, sleep

def rg_computer(si, pn, queues):
    import molass_legacy.KekLib, SerialAnalyzer
    from molass_legacy._MOLASS.SerialSettings import initialize_settings
    from .SharedInfo import QUEUE_RG
    from molass_legacy.GuinierAnalyzer.SimpleGuinier import SimpleGuinier
    initialize_settings()

    print("rg_computer", si)
    rg_queue = queues[QUEUE_RG]

    while True:
        i = rg_queue.get()
        if i < 0:
            if i == -1:
                t0 = time()
                si.get_buffer_ready()
            elif i == -2:
                print("It took %.3g seconds" % (time()-t0))
            else:
                break
            continue

        data = si.get_xr_data(i)
        sg = SimpleGuinier(data)
        # print([i], "rg computed", sg.Rg)

    print("rg_computer stop")



"""
    UnifiedDecompResult.py

    Copyright (c) 2018-2024, SAXS Team, KEK-PF
"""
import copy
import numpy as np
from molass_legacy.KekLib.BasicUtils import Struct
import molass_legacy.KekLib.DebugPlot as plt
from OptRecsUtils import compute_area_proportions

SUFFICIENT_TOP_RATIO    = 0.3
SELECTABLE_TOP_RATIO    = 0.085 # < 0.086 for 20161202,  > 0.082 for OAGIwyatt_02, < 0.11 for OA_Ald
DEPENDENT_DIFF_RATIO    = 0.02  # > 0.006 for the 1st minor peak of 20181203, > 0.034 for Dec02, < 0.026 for Dec02
DEPENDENT_MAJOR_RATIO   = 0.5   # < 0.89 for the 1st minor peak of 20181203, < 0.57 for monoE
DEPENDENT_LARGER_RATIO  = 0.25  # < 0.28 for OA_Ald
FWHM_SPAN_RATIO         = 1.3   # < 1.34 for Mar01
AREA_PROPORTION_LIMIT   = 0.03  # > 0.0096 for 2-th peak HasA(EGH)
IGNORABLE_PROPORTION    = 0.03

class UnifiedDecompResult:
    def __init__(self,
                    xray_to_uv=None,
                    x_curve=None, x=None, y=None,
                    opt_recs=None,
                    max_y_xray = None,
                    model_name=None,
                    decomposer=None,
                    uv_y=None,
                    opt_recs_uv=None,
                    max_y_uv = None,
                    nresid_uv=None,
                    global_flag=False,
                    debug_info=None,
                    ): 
        self.xray_to_uv = xray_to_uv
        self.x_curve = x_curve
        self.x = x
        self.y = y
        self.opt_recs = opt_recs
        self.max_y_xray = max_y_xray
        self.model_name = model_name
        self.decomposer = decomposer
        self.uv_y = uv_y
        self.opt_recs_uv = opt_recs_uv
        self.max_y_uv = max_y_uv
        self.debug_info = debug_info
        self.nresid_uv = None if uv_y is None else self.compute_normalized_resid()
        self.global_flag = global_flag
        self.marked_as_dependent = np.zeros(len(opt_recs))

    def compute_normalized_resid(self):
        resid_y = copy.deepcopy(self.uv_y)

        for rec in self.opt_recs_uv:
            func = rec[1]
            resid_y -= func(self.x)

        nresid = np.average(np.abs(resid_y)) / self.max_y_uv

        if self.debug_info is not None:
            print("nresid=", nresid)
            x = self.x
            uv_y = self.uv_y
            with plt.Dp():
                fig, ax = plt.subplots()
                ax.set_title(self.debug_info[0])
                ax.plot(x, uv_y)
                for rec in self.opt_recs_uv:
                    func = rec[1]
                    cy = func(x)
                    ax.plot(x, cy, ":")
                ax.plot(x, resid_y, ":", color="red")
                fig.tight_layout()
                plt.show()

        return nresid

    def set_area_proportions(self):
        def set_area_proportions_impl(x, recs):
            areas = []
            for k, rec in enumerate(recs):
                cy = rec.evaluator(x)
                areas.append(np.sum(cy))
            props = np.array(areas)/np.sum(areas)
            for rec, p in zip(recs, props):
                rec.peak.area_prop = p

        set_area_proportions_impl(self.x, self.opt_recs)
        set_area_proportions_impl(self.x, self.opt_recs_uv)

    def get_range_edit_info(self, logger=None, debug=False):
        if debug:
            from importlib import reload
            import Decomposer.OptRecsUtils
            reload(Decomposer.OptRecsUtils)

        opt_recs = self.opt_recs
        x = self.x

        if debug:
            from molass_legacy.Decomposer.OptRecsUtils import debug_plot_opt_recs
            debug_plot_opt_recs(self.x_curve, opt_recs, eval_x=x, title="get_range_edit_info entry opt_recs")

        select_matrix = []
        editor_ranges = []
        top_x_list = []

        num_eltns = len(opt_recs)
        size_x = len(x)
        for k, rec in enumerate(opt_recs):
            ee_selection = [0] * num_eltns
            ee_selection[k] = 1
            select_matrix.append(ee_selection)

            range_list = rec.get_range_list(x)

            editor_ranges.append(range_list)
            peak = rec[3]

            # - self.x[0] is needed for Stochastic Models while self.x[0] == 0 for traditional models
            top_x_list.append( peak.top_x - self.x[0] )

        if debug:
            print('select_matrix(before update)=', select_matrix)

        self.update_all_selections(select_matrix, editor_ranges, top_x_list, logger, debug=debug)

        if debug:
            from molass_legacy.Decomposer.OptRecsUtils import debug_plot_opt_recs_impl
            print('select_matrix(after update)=', select_matrix)
            y = self.y
            with plt.Dp():
                fig, ax = plt.subplots()
                ax.set_title("get_range_edit_info exit opt_recs with editor_ranges")
                debug_plot_opt_recs_impl(ax, x, y, opt_recs)
                for range_list in editor_ranges:
                    for f, t in range_list:
                        ax.axvspan(x[f], x[t], alpha=0.3, color="cyan")
                fig.tight_layout()
                plt.show()

        return Struct(
                    select_matrix=select_matrix,
                    editor_ranges=editor_ranges,
                    top_x_list=top_x_list,
                    )

    def update_all_selections(self, select_matrix, editor_ranges, top_x_list, logger, debug=False):
        opt_recs = self.opt_recs
        max_y = self.max_y_xray

        for k, rec in enumerate(opt_recs):
            if self.is_dependent_peak(k, top_x_list, logger, debug):
                if debug:
                    print([k], 'updating for the dependent peak')
                self.update_one_selection(0, select_matrix, k)
                editor_ranges[k].clear()
                self.marked_as_dependent[k] = 1

    def update_one_selection(self, num_ranges, select_matrix, k, ignorable_flags=None):
        nearest_k = self.get_nearest_major_peak_index(k, ignorable_flags)
        if nearest_k is None:
            # TODO: pH6
            return

        major_selection = select_matrix[nearest_k]
        select_list = select_matrix[k]
        if num_ranges == 0:
            select_list[k] = 0
            major_selection[k] = 1
        else:
            select_list[k] = 1
            major_selection[k] = 0

    def get_nearest_major_peak_index(self, this_k, ignorable_flags=None):
        opt_recs = self.opt_recs

        nearest_k = None
        nearest_diff = None
        this_x = opt_recs[this_k][3].top_x
        for k, rec in enumerate(opt_recs):
            if k == this_k or rec[0] < 0 or (ignorable_flags is not None and ignorable_flags[k] == 1):
                continue
            other_x = rec[3].top_x
            diff = abs(this_x - other_x)
            if nearest_diff is None or diff < nearest_diff:
                nearest_diff = diff
                nearest_k = k
        return nearest_k

    def is_dependent_peak(self, k, top_x_list, logger, debug=False):
        opt_recs = self.opt_recs
        max_y = self.max_y_xray

        rec = opt_recs[k]
        rec_id = rec[0]
        if rec_id >= 0:
            # major peak should be indepenent
            return False

        minor_rec = rec
        peak = minor_rec[3]
        if peak.sign == -1:
            # negative peak must be dependent
            return True

        top_x = top_x_list[k]
        func = minor_rec[1]
        top_y = func(top_x)
        top_ratio = top_y/max_y
        if debug:
            print( [k], 'top_ratio=', top_ratio, 'top_y=', top_y, 'max_y=', max_y )
        top_dependence = top_ratio < SELECTABLE_TOP_RATIO

        minor_peak = minor_rec[3]
        minor_func = minor_rec[1]
        major_k = self.get_nearest_major_peak_index(k)
        if major_k is None:
            # occured in pH6
            return False

        major_rec = opt_recs[major_k]
        major_peak = major_rec[3]
        major_top_x = major_peak.top_x
        top_dist = abs(top_x - major_top_x)
        span_ratio = top_dist/major_peak.fwhm
        if debug:
            print( [k], 'top_dist=', top_dist, 'major_peak.fwhm=', major_peak.fwhm, 'span_ratio=', span_ratio )
        if span_ratio > FWHM_SPAN_RATIO:
            # too distant peak cannot be dependent
            return False

        major_func = major_rec[1]
        # print('minor_func=', minor_func)
        # print('minor_func.dict_params=', minor_func.dict_params)
        x = np.arange(minor_peak.flimL, minor_peak.flimR+1)
        y1 = major_func(x)
        y2 = minor_func(x)
        upper_y = np.max([y1, y2], axis=0)
        total_y = y1 + y2
        diff_y = np.abs(total_y - upper_y)
        denom = np.sum(upper_y)
        diff_ratio = abs(np.sum(diff_y)/denom)
        major_ratio = abs(np.sum(y1)/denom)
        sub_y = y2 - y1
        larger_ratio = np.sum(sub_y[sub_y > 0])/abs(np.sum(y2))

        if debug:
            print('top_x=', top_x, 'max_y=', max_y, 'top_ratio=', top_ratio)
            print('diff_ratio=', diff_ratio, 'major_ratio=', major_ratio)
            print('larger_ratio=', larger_ratio)
            with plt.Dp():
                fig, ax = plt.subplots()
                ax.set_title(str(minor_peak))
                ax.plot(x, total_y, label='total_y', linewidth=3, alpha=0.3)
                ax.plot(x, y1, label='major_y', linewidth=3, alpha=0.3)
                ax.plot(x, y2, label='minor_y', linewidth=3, alpha=0.3)
                ax.plot(x, upper_y, ':', label='upper_y', linewidth=3)
                ax.plot(x, diff_y, ':', label='diff_y', linewidth=3)
                ax.legend()
                fig.tight_layout()
                plt.show()

        if larger_ratio > DEPENDENT_LARGER_RATIO:
            return False

        # check these ratios before checking the top_ratio
        shape_dependence = ( diff_ratio < DEPENDENT_DIFF_RATIO and major_ratio > DEPENDENT_MAJOR_RATIO )

        if top_dependence == shape_dependence:
            return top_dependence

        if shape_dependence:
            if top_ratio < SUFFICIENT_TOP_RATIO:
                from molass_legacy.KekLib.BasicUtils import ordinal_str
                logger.info( '%s peak has been judged as dependent with diff_ratio=%.3g' % ( ordinal_str(k+1), diff_ratio) )
                return shape_dependence

        return top_dependence

    def identify_ignorable_elements(self):
        if self.global_flag:
            return np.zeros(len(self.opt_recs))

        assert len(self.opt_recs) == len(self.opt_recs_uv)

        flags = []
        for k, p in enumerate(self.proportions):
            flag = ( p < IGNORABLE_PROPORTION
                    or self.marked_as_dependent[k] == 1)
            flags.append(flag)

        return np.array(flags)

    def compute_element_score(self, num_elements):
        count = num_elements
        for rec in self.opt_recs_uv:
            if rec.peak.sign < 0:
                count += 1
        return self.nresid_uv * count

    def compute_synthetic_score(self, ignorable_flags=None, select_matrix=None, debug=False):
        if ignorable_flags is None:
            ignorable_flags = self.identify_ignorable_elements()
        if select_matrix is None:
            info = self.get_range_edit_info()
            select_matrix = info.select_matrix

        if debug:
            print('self.nresid_uv=', self.nresid_uv)

        scores = []
        for k in range(len(self.opt_recs)):
            if ignorable_flags[k] == 0:
                ee_select = select_matrix[k]
                # self.logger.info('select_matrix[%d]=%s' % (k, str(ee_select)))
                scores.append( self.compute_element_score(np.sum(ee_select)) )
        score = np.average(scores)

        if debug:
            print('score=', score)

        return score

    def remove_unwanted_elements(self):
        import logging
        logger = logging.getLogger(__name__)

        self.proportions = compute_area_proportions(self.x, self.opt_recs)
        logger.info("debug: decomposed proportions are %s", str(self.proportions))
        to_be_removed = np.where(self.proportions < AREA_PROPORTION_LIMIT)[0]
        if len(to_be_removed) == 0:
            # nothing to remove
            return

        logger.info("decomposed proportions are %s", str(self.proportions))

        n = len(self.opt_recs)

        for k in to_be_removed[::-1]:
            del self.opt_recs[k]
            del self.opt_recs_uv[k]

        # recompute proportions
        # calling self.set_proportions() instead here results in a strange behavior. why?
        self.proportions = compute_area_proportions(self.x, self.opt_recs)

        logger.info("unwanted small-proportion %s-th elements have been removed from decomposition results.", str(to_be_removed))
        logger.info("recomputed proportions are %s", str(self.proportions))

    def set_proportions(self):
        self.proportions = compute_area_proportions(self.x, self.opt_recs)


# coding: utf-8
"""

    Optimize.ModelParams.py

    Copyright (c) 2017-2021, SAXS Team, KEK-PF

"""
import cupy as np

def gaussian(x, h, m, s):
    return h * np.exp(-((x-m)/s)**2)

def egh(x, H=1, tR=0, sigma=1.0, tau=0):
    x_  = x - tR
    s2  = 2 * sigma**2
    z   = s2 + tau*x_
    z_neg   = z <= 0
    z_pos   = z > 0

    zero_part = np.zeros( len(x) )[z_neg]
    posi_part = H * np.exp( -x_[z_pos]**2/( s2 + tau*x_[z_pos] ) )

    if tau > 0:
        parts = [ zero_part, posi_part ]
    else:
        parts = [ posi_part, zero_part ]

    return np.hstack( parts )



"""
    RgVisibleFigure.py

    Copyright (c) 2022-2024, SAXS Team, KEK-PF
"""
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from molass_legacy.KekLib.OurTkinter import Tk, Dialog
from molass_legacy.KekLib.OurMatplotlib import NavigationToolbar
from molass_legacy.SerialAnalyzer.DataUtils import get_in_folder
from GuinierTools.RgCurveUtils import compute_rg_curves
from .FvScoreConverter import convert_score
from .LrfExporter import LrfExporter

DX_RATIO = 0.05
DY_RATIO = 0.04
SQ_DIST = DX_RATIO**2 + DY_RATIO**2     # squared distance in normalized space

class RgVisibleFigure(Dialog):
    def __init__(self, parent, js_canvas, with_range=False):
        self.parent = parent
        self.js_canvas = js_canvas
        self.with_range = with_range
        self.elution_model = js_canvas.elution_model
        self.fullopt = js_canvas.fullopt
        self.class_code = js_canvas.dialog.class_code
        self.params = js_canvas.get_current_params()
        self.separate_params = self.fullopt.split_params_simple(self.params)

        self.fv = convert_score(js_canvas.get_current_fv())
        # exporter = self.fullopt.params_type.get_exporter(js_canvas.optinit_info.sd, js_canvas.dsets, self.fullopt, self.params)
        exporter = LrfExporter(self.fullopt, self.params, js_canvas.dsets)
        # exporter.compute_LRF_xr()
        self.peaktops = exporter.xr_result.peaktops
        Dialog.__init__(self, parent, "Rg-visible Figure", visible=False)

    def get_secparams(self):
        secparams = self.separate_params[7]
        if self.elution_model == 1:
            t0, rp, N, me, T, mp = secparams
            P = N*T
            m = me+mp
        else:
            t0, P, rp, m = secparams[0:4]
        return t0, P, rp, m

    def show(self):
        self._show()

    def body(self, body_frame):
        cframe = Tk.Frame(body_frame)
        cframe.pack()
        tframe = Tk.Frame(body_frame)
        tframe.pack(fill=Tk.X)

        fig_width = 6
        if self.with_range:
            t0, P, rp, m = self.get_secparams()
            self.sec_range = t0, t0+P
            x = self.fullopt.xr_curve.x
            width = max(t0+P, x[-1]) - min(t0, x[0])
            ratio = width/len(x)
            fig_width = min(18, fig_width*ratio)

        fig, ax = plt.subplots(figsize=(fig_width,5))
        self.fig = fig
        axt = ax.twinx()
        axt.grid(False)
        self.axes = (ax, axt)
        self.mpl_canvas = FigureCanvasTkAgg(self.fig, cframe)
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack(fill=Tk.BOTH, expand=1)
        self.toolbar = NavigationToolbar(self.mpl_canvas, tframe)
        self.toolbar.update()

        self.draw_figure(fig, ax, axt)

    def draw_figure(self, fig, ax, axt):
        in_folder = get_in_folder()
        nc = self.fullopt.n_components - 1
        job_name = self.js_canvas.dialog.get_job_info()[0]
        ax.set_title("%s, NC=%d, F=%s, SV=%.3g, J=%s" % (in_folder, nc, self.class_code, self.fv, job_name), fontsize=14, y=1.05)
        ax.set_xlabel("Time (Eno)")
        ax.set_ylabel("Intensity")
        axt.set_ylabel("$R_g$")

        fullopt = self.fullopt

        x = fullopt.xr_curve.x
        y = fullopt.xr_curve.y
        lrf_info = fullopt.objective_func(self.params, return_lrf_info=True)
        xr_cy_list = lrf_info.get_xr_cy_list()
        xr_ty = lrf_info.xr_ty

        ax.plot(x, y, color="orange")

        stop = -2 if fullopt.separate_eoii else -1
        for k, cy in enumerate(xr_cy_list[0:stop]):
            ax.plot(x, cy, ":")

        if fullopt.separate_eoii:
            cy = xr_cy_list[-1]
            ax.plot(x, cy, ":", color="pink")

        cy = xr_cy_list[stop]
        ax.plot(x, cy, color="red")

        ax.plot(x, xr_ty, ":", color="red")

        xr_params, _, rg_params = self.separate_params[0:3]

        ymin, ymax = ax.get_ylim()
        ax.set_ylim(ymin, ymax)

        if self.with_range:
            xmin, xmax = ax.get_xlim()
            dx = (xmax - xmin)*0.02
            k = 0
            for t_, color in zip(self.sec_range, ["red", "green"]):
                ax.plot([t_, t_], [ymin, ymax], ":", color=color, lw=3)
                if k == 0:
                    ax.text(t_+dx, ymin, "$t_0$", fontsize=20, color=color, ha="left", va="bottom")
                else:
                    ax.text(t_-dx, ymin, "$t_0 + P$", fontsize=20, color=color, ha="right", va="bottom")
                k += 1

        xmin, xmax = ax.get_xlim()
        ax.set_xlim(xmin, xmax)

        xsize = (xmax - xmin)
        ysize = (ymax - ymin)
        dx = xsize*0.02
        dy = ysize*0.02

        n = len(self.peaktops)
        m, r = divmod(n, 2)

        text_pos = []
        def find_better_pos(tx, ty):
            for x, y in text_pos:
                dist2 = ((tx - x)/xsize)**2 + ((ty - y)/ysize)**2
                if dist2 < SQ_DIST:
                    tx += np.sign(tx - x) * xsize * DX_RATIO
                    ty += np.sign(ty - y) * ysize * DY_RATIO
                    break
            return tx, ty

        for k, (px, py) in enumerate(self.peaktops):
            if px < 0:
                continue

            ax.plot(px, py, "o", color="yellow")
            dy_ = dy
            if k <= m:
                if r == 1 and k == m:
                    ha = "center"
                    dx_ = 0
                else:
                    ha = "right"
                    dx_ = -dx
            else:
                ha = "left"
                dx_ = dx
            tx = px+dx_
            ty = py+dy_
            tx, ty = find_better_pos(tx, ty)
            ax.text(tx, ty, "%.3g" % rg_params[k], ha=ha, fontsize=16, alpha=0.5)
            text_pos.append((px, py))
            text_pos.append((tx, ty))

        # Rg curves
        if self.with_range:
            from SecTheory.ConformanceDemo import SecConfCurve
            t0, P, rp, m = self.get_secparams()
            secconf_curve = SecConfCurve(t0, P, rp, m, size=100)
            x_ = x[x > t0]
            y_ = secconf_curve(x_)
            axt.plot(x_, y_, ":", color="purple", lw=3, label="conformance curve")

        xr_ty_ = xr_ty - xr_cy_list[-1]      # remove baseline part
        if len(xr_params.shape) == 2:
            xr_heights = xr_params[:,0]
        else:
            xr_heights = xr_params
        rg_curves1, rg_curves2 = compute_rg_curves(x, xr_heights/np.max(xr_heights), rg_params, xr_cy_list[:-1], xr_ty_, self.fullopt.rg_curve)

        k = 0
        for (x1, rg1), (x2, rg2) in zip(rg_curves1, rg_curves2):
            axt.plot(x1, rg1, color='gray', alpha=0.5)
            axt.plot(x2, rg2, ':', color='black')
            k += 1

        ymin, ymax = axt.get_ylim()
        axt.set_ylim(min(10, ymin), max(50, ymax))
        # axt.legend(loc='upper right', fontsize=16)

        fig.tight_layout()
        fig.subplots_adjust(top=0.88)
        self.mpl_canvas.draw()



"""
    Selective/MinimizerUtils.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import logging
import numpy as np
from scipy.stats import linregress
from molass_legacy.DataStructure.SvdDenoise import get_denoised_data
from molass_legacy.GuinierAnalyzer.SimpleGuinier import SimpleGuinier
from molass_legacy.GuinierAnalyzer.SimpleGuinierScore import compute_rg
from molass_legacy.DataStructure.AnalysisRangeInfo import shift_range_from_to_by_x
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.Models.ModelUtils import compute_cy_list

class RgComputer:
    def __init__(self, model, D, E, qv, x, y, compute_range, num_components, target_i):
        self.logger = logging.getLogger(__name__)
        self.logger.info("RgComputer has been created with model=%s, num_components=%d, target_i=%d" % (model.get_name(), num_components, target_i))
        self.model = model
        self.num_components = num_components
        self.M = get_denoised_data(D, rank=num_components)
        self.E = E
        self.qv = qv
        self.x = x
        self.y = y
        self.compute_range = compute_range
        self.target_i = target_i

    def compute_C(self, params_array):
        C = np.array(compute_cy_list(self.model, self.x, params_array))
        return C

    def compute_rg_list(self, C, return_sg=False, return_nI=False,
                        return_with_pos=False, discard_bad_rg=False,
                        gslice=None, debug=False):
        # C
        # P = MC⁺
        model = self.model
        M = self.M
        E = self.E
        qv = self.qv
        x = self.x
        compute_range = self.compute_range
        target_i = self.target_i
        
        if debug:
            y = self.y
            with plt.Dp():
                fig, ax = plt.subplots()
                ax.set_title("compute_rg_list: C")
                ax.plot(x, y, color="orange")
                for k, cy in enumerate(C):
                    ax.plot(x, cy, ":", label="component-%d" % k)
                for k, (f, t) in enumerate(compute_range.get_fromto_list()):
                    print([k], f, t)
                    ax.axvspan(f, t, alpha=0.3, color="C%d" % k)
                ax.legend()
                fig.tight_layout()
                plt.show()

        if return_sg:
            sg_list = []
        else:
            assert gslice is not None
            qv2 = qv[gslice]**2

        rg_list = []
        if return_nI:
            assert gslice is not None
            nI_list = []
        for f, t in compute_range.get_fromto_list():
            f, t = shift_range_from_to_by_x(x, f, t)
            slice_ = slice(f,t+1)
            M_ = M[:,slice_]
            C_ = C[:,slice_]
            E_ = E[:,slice_]
            P_ = M_ @ np.linalg.pinv(C_)
            if return_sg:
                Minv = np.linalg.pinv(M_)
                W = Minv @ P_
                Pe = np.sqrt(E_**2 @ W**2)
                data = np.array([qv, P_[:,target_i], Pe[:,target_i]]).T
                if debug:
                    gslice = slice(0, len(qv)//4)
                    qv_ = qv[gslice]
                    iv_ = P_[gslice,target_i]
                    with plt.Dp():
                        fig, (ax1,ax2) = plt.subplots(ncols=2, figsize=(12,5))
                        ax1.set_title("Log Plot in %s" % ((f,t),))
                        ax1.set_yscale("log")
                        ax1.plot(qv_, iv_)
                        ax2.set_title("Guinier Plot in %s" % ((f,t),))
                        ax2.plot(qv_**2, np.log(iv_))
                        fig.tight_layout()
                        plt.show()
                sg = SimpleGuinier(data)
                if discard_bad_rg:
                    if sg.Rg is None or sg.Rg < 10:
                        continue
                sg_list.append(sg)
                rg = sg.Rg
            else:
                slope, intercept = linregress(qv2, np.log(P_[gslice,target_i]))[0:2]
                rg = compute_rg(slope)
                if return_nI:
                    nI_list.append(P_[:,target_i]/np.exp(intercept))
            if return_with_pos:
                pos = (t+f)/2   # be aware that this is a "shifted" position
                rg_list.append((pos, rg))
            else:
                rg_list.append(rg)

        if return_sg:
            return rg_list, sg_list
        else:
            if return_nI:
                return rg_list, nI_list
            else:
                return rg_list



"""
    EoiiCorrector.py

    Copyright (c) 2023, SAXS Team, KEK-PF
"""
import logging
import numpy as np
from bisect import bisect_right
from scipy.optimize import minimize
from SvdDenoise import get_denoised_data
import molass_legacy.KekLib.DebugPlot as plt

class EoiiCorrector:
    def __init__(self, sd, paired_ranges, debug=False):
        if debug:
            from importlib import reload
            import LRF.PnoScdMap
            reload(LRF.PnoScdMap)
        from .PnoScdMap import PnoScdMap

        self.logger = logging.getLogger(__name__)
        self.logger.info("paired_ranges=%s", str(paired_ranges))
        D, E, qv, ecurve = sd.get_xr_data_separate_ly()
        self.pre_recog = sd.pre_recog
        self.D = D
        self.E = E
        self.qv = qv
        self.i = sd.xray_index
        self.x = ecurve.x
        self.y = ecurve.y
        self.paired_ranges = paired_ranges
        self.pno_map = PnoScdMap(sd, paired_ranges)

    def correct_params_list(self, func, params_list, debug=False):

        ret_params_list = []
        corrected = False
        for k, params in enumerate(params_list):
            color = self.pno_map.get_color(k)
            print([k], "color=", color)
            if color in ["yellow", "red"]:
                params = self.compute_corrected_cy(func, params, self.paired_ranges[k])
                corrected = True
            ret_params_list.append(params)

        if debug:
            x = self.x
            y = self.y

            with plt.Dp():
                fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12,5))
                fig.suptitle("EoiiCorrector correct_params_list debug")
                for ax, p_list in (ax1, params_list),(ax2, ret_params_list):
                    ax.plot(x, y)
                    for k, params in enumerate(p_list):
                        print([k], params)
                        ax.plot(x, func(x, *params), ":")
                fig.tight_layout()
                plt.show()

        return corrected, ret_params_list

    def compute_corrected_cy(self, func, params, prange, debug=False):
        f, t = prange.get_concatenated_range()
        print("prange=", prange, (f, t))
        return compute_corrected_cy_impl(self.qv, self.D, self.E, self.x, self.y, self.i, f, t, func, params, logger=self.logger, debug=debug)

USE_BOUNDED_LRF = False

def compute_corrected_cy_impl(qv, inD, inE, x, y, i, f, t, func, params, logger=None, debug=False):
    if USE_BOUNDED_LRF:
        if debug:
            from importlib import reload
            import BoundedLRF.BoundedLrfSolver
            reload(BoundedLRF.BoundedLrfSolver)
        from BoundedLRF.BoundedLrfSolver import BoundedLrfSolver
    else:
        if debug:
            from importlib import reload
            import LRF.LrfRgUtils
            reload(LRF.LrfRgUtils)
        from .LrfRgUtils import compute_rg_from_qvDEP

    slice_ = slice(f, t+1)
    x_ = x[slice_]
    target_y = func(x, *params)
    # target_y[slice_] = y[slice_]

    def objective(p):
        h, m, s, t, a, aqi, bqi = p
        c = func(x, h, m, s, t, a)
        return np.linalg.norm(aqi*c + bqi*c**2 - target_y)

    D = inD[:,slice_]
    E = inE[:,slice_]

    params_ = params

    results = []
    num_iter = 1
    for k in range(num_iter + 1):
        cy = func(x_, *params_)
        C = np.array([cy, cy**2])
        try:
            if USE_BOUNDED_LRF:
                solver = BoundedLrfSolver(qv, D, E, C=C, i=i)
                P_, C__, Rg, R_, L_, hK, hL, bq_bounds_, coerced_bq_ = solver.solve()
                aq_, bq_ = P_.T
            else:
                Cinv = np.linalg.pinv(C)
                P_ = D @ Cinv
                aq_, bq_ = P_.T
                Rg = compute_rg_from_qvDEP(qv, D, E, P_)
            results.append((cy, aq_, bq_, params_, Rg))
            if k < num_iter:
                aqi = aq_[i]
                bqi = bq_[i]
                ret = minimize(objective, (*params_, aqi, bqi))
                params_ = ret.x[0:5]
        except:
            from molass_legacy.KekLib.ExceptionTracebacker import log_exception
            log_exception(logger, "compute_corrected_cy: ")
            break

    if debug:
        cy, aq_, bq_, params_, Rq  = results[0]
        ig = bisect_right(qv, 1.8/Rg)
        gslice = slice(0,ig)
        gqv = qv[gslice]
        gqv2 = gqv**2
        j = np.argmax(cy)
        gy = D[gslice,j]

        with plt.Dp():
            fig, (ax1, ax2, ax3, ax4) = plt.subplots(ncols=4, figsize=(20,5))
            fig.suptitle("EoiiCorrector compute_corrected_cy debug", fontsize=20)
            ax1.plot(x, y, color="gray", alpha=0.5)
            # ax1.plot(x_, y_)
            # ax1.plot(x, func(x, *params), ":", label="init estimation")
            ymin, ymax = ax1.get_ylim()
            ax1.set_ylim(ymin, ymax)
            for j in f, t:
                ax1.plot([j, j], [ymin, ymax], ":", color="gray")

            ax4.plot(gqv2, np.log(gy), color="gray", alpha=0.5, label="apparent data")

            for k, (cy_, aq_, bq_, params_, Rq) in enumerate(results):
                j = np.argmax(cy_)
                cj = cy_[j]

                cy = func(x, *params_)
                state_label = "init estimation" if k == 0 else "corrected"
                ax1.plot(x, cy, ":", label="[%d] %s" % (k, state_label), lw=2)

                ax2.plot(qv, aq_*cj, label="[%d] $A_q, \;(R_g=%.3g)$" % (k, Rg), alpha=0.5)
                ax3.plot(qv, bq_*cj**2, label="[%d] $B_q, \; (R_g=%.3g)$" % (k, Rg), alpha=0.5)
                ax4.plot(gqv2, np.log(aq_[gslice]*cj), label="[%d] $A_q, \;(R_g=%.3g)$" % (k, Rg), alpha=0.5)

            for ax in ax1, ax2, ax3, ax4:
                ax.legend()

            fig.tight_layout()
            plt.show()

    return params_



# coding: utf-8
"""
    SimpleGuinierFolderAnalyzer.py

    Copyright (c) 2017-2019, SAXS Team, KEK-PF
"""
import os
import re
import numpy as np
import glob
from molass_legacy.KekLib.NumpyUtils                 import np_loadtxt
from molass_legacy.KekLib.OurTkinter                 import Tk, Font, is_empty_val
from molass_legacy.KekLib.TkUtils                    import adjusted_geometry, split_geometry
from molass_legacy.KekLib.TkCustomWidgets            import FileEntry, FolderEntry
from CanvasFrame                import CanvasFrame
from molass_legacy.KekLib.OurMatplotlib              import NavigationToolbar
from CanvasDialog               import CanvasDialog
from DataUtils                  import get_pytools_folder
from SerialData                 import SerialData, find_conc_files
from molass_legacy._MOLASS.SerialSettings             import get_setting
from SimpleGuinierGuiUtils      import window_title, CheckbuttonFrame, TypeInfoPanel

NUM_EXTRAS  = 10
USE_SAVED_DATA  = True
if USE_SAVED_DATA:
    from SerialDataLoader       import SerialDataLoader

class SimpleGuinierFolderAnalyzer( Tk.Toplevel ):
    def __init__( self, parent, folder_no, folder_path, folder_info, x, z, r, type_names, report_folder=None ):
        self.parent = parent
        self.folder_no = folder_no
        self.folder_path = folder_path
        self.folder_info = folder_info
        if report_folder is None:
            report_folder = self.parent.report_folder.get() + '/analysis-%03d' % folder_no
        self.report_folder = report_folder
        self.num_zx_results = 0
        for left, peak, right in folder_info:
            self.num_zx_results += right - left + 2 + NUM_EXTRAS    # left-peak, peak-right
        self.num_sg_results = len(x) - self.num_zx_results
        if USE_SAVED_DATA:
            self.loader = SerialDataLoader()
            in_folder = report_folder + '/averaged'
            self.loader.load_xray_data_in_another_thread( in_folder )
        else:
            conc_folder, conc_file = self.get_conc_file( folder_path )
            self.serial_data = SerialData( conc_folder, folder_path, conc_file=conc_file )
        self.averaged_intensity_array = None
        self.init_x  = x
        self.init_z  = z
        self.init_r  = r
        self.type_names = type_names
        temp_dict = {}
        for v in z:
            temp_dict[v] = 1
        self.types = sorted( temp_dict.keys() )
        Tk.Toplevel.__init__( self, parent )
        self.title( window_title )

        frame1 = Tk.Frame( self )
        frame1.pack()

        self.canvas = CanvasFrame( frame1, figsize=(12, 8) )
        self.canvas.pack( side=Tk.LEFT )

        self.toolbar = NavigationToolbar( self.canvas.mpl_canvas, self.canvas )
        self.toolbar.update()
        self.create_popup_menu()

        frame12 = Tk.Frame( frame1 )
        frame12.pack( side=Tk.RIGHT, fill=Tk.BOTH  )
        label = Tk.Label( frame12, text='Control Panel' )
        label.pack( anchor=Tk.N )

        self.cb_frame = CheckbuttonFrame( frame12, types=self.types )
        self.cb_frame.pack()
        self.update_xzr( force=True )

        redraw_btn   = Tk.Button( frame12, text='redraw', command=self.redraw )
        redraw_btn.pack()

        typeinfo_btn   = Tk.Button( frame12, text='type help', command=self.show_type_info )
        typeinfo_btn.pack()

        if False:
            print( 'len(x)=', len(x) )
            print( 'x=', x )
            print( 'num_sg_results=', self.num_sg_results )
            print( 'folder_info=', self.folder_info )

        self.range_indicators = []
        for left, peak, right in self.folder_info:
            self.range_indicators.append( left )
            self.range_indicators.append( peak )
            self.range_indicators.append( right )

        zx_base = self.num_sg_results + 1       # there is one space before extrapolation starts
        for left, peak, right in self.folder_info:
            self.range_indicators.append( zx_base )
            zx_peak = zx_base + ( peak - left ) + 5
            self.range_indicators.append( zx_peak )
            zx_base = zx_peak + 2
            self.range_indicators.append( zx_base )
            zx_right = zx_base + ( right - peak ) + 5 
            self.range_indicators.append( zx_right )
            zx_base = zx_right + 2

    def get_conc_file( self, folder_path ):
        datafile_recs  = list( find_conc_files( folder_path ) )
        if len(datafile_recs) > 0:
            conc_file = datafile_recs[0][1]
            conc_folder = folder_path
        else:
            uv_folder_path = folder_path.replace( 'Backsub', 'spectra' )
            datafile_recs  = list( find_conc_files( uv_folder_path ) )
            if len(datafile_recs) > 0:
                conc_file = datafile_recs[0][1]
                conc_folder = folder_path
            else:
                # TODO:
                assert False
        return conc_folder, conc_file

    def show( self ):
        self.canvas.draw( self.draw )
        self.geometry( adjusted_geometry(self.geometry()) )
        self.fig.canvas.mpl_connect( 'button_press_event', self.on_button_press )

    def create_popup_menu( self ):
        self.popup_menu = Tk.Menu( self, tearoff=0 )
        self.popup_menu.add_command( label='Animation', command=self.do_animation )

    def update_xzr( self, force=False ):
        if force or self.cb_frame.changed:
            index   = self.cb_frame.make_index( self.init_z )
            self.x  = self.init_x[index]
            self.z  = self.init_z[index]
            self.r  = self.init_r[index]

    def draw( self, fig ):
        self.fig = fig
        ax = fig.add_subplot( 111 )
        self.axes = [ax]
        self.plot( ax, get_lim=True )
        fig.tight_layout()

    def plot( self, ax, draw_flags=None, get_lim=False ):
        folderno = '' if self.folder_no < 0 else 'Foder No: %03d      ' % self.folder_no
        ax.set_title( folderno + 'Path: %s' % ( self.folder_path ) )
        ax.set_xlabel( 'File No' )
        ax.set_ylabel( 'Rg' )

        for i, n in enumerate(self.types):
            if draw_flags is not None:
                if draw_flags[i] == 0:
                    continue

            index = self.z == n
            x_ = self.x[index]
            if len( x_ ) == 0:
                continue
            r_ = self.r[index]
            ax.plot( x_, r_, 'o', markersize=3, label='type %d' % n )

        if get_lim:
            self.xlim = np.array( ax.get_xlim() )   # deep copy
            self.ylim = np.array( ax.get_ylim() )   # deep copy

        ax.set_xlim( self.xlim )
        ax.set_ylim( self.ylim )

        for x in self.range_indicators:
            ax.plot( [ x, x ], self.ylim, color='green', linewidth=1, alpha=0.2 )

        ax.legend()

    def redraw( self ):
        self.update_xzr()

        ax = self.axes[0]
        ax.cla()
        # draw_flags = [ var.get() for var in self.cb_frame.cb_vars ]
        def redraw_closure( fig ):
            self.plot( ax )
            ax.set_xlim( self.xlim )
            ax.set_ylim( self.ylim )

        self.canvas.draw( redraw_closure )

    def on_button_press( self, event ):
        if event.button == 3:
            self.on_right_button_press( event )
            return

    def on_right_button_press( self, event ):
        print( 'on_right_button_press' )

        if event.inaxes not in self.axes:
            return

        # https://stackoverflow.com/questions/17711099/programmatically-change-matplotlib-toolbar-mode-in-qt4
        if self.toolbar._active == 'ZOOM':
            # toggle zoom mode
            self.toolbar.zoom()

        print( 'x=', event.xdata )
        print( 'y=', event.ydata )

        dist = ( self.x - event.xdata )**2 + ( self.r - event.ydata )**2
        m = np.argmin( dist )
        self.file_index = m

        w, h, x, y = split_geometry( self.canvas.mpl_canvas_widget.winfo_geometry() )
        h_ = h
        # print( 'self.mpl_canvas_widget.winfo_geometry()=', (w, h, x, y) )
        w, h, x, y = split_geometry( self.winfo_geometry() )
        # print( 'self.winfo_geometry()=', (w, h, x, y) )
        # TODO: better way to get the mouse cursor position
        self.popup_menu.entryconfig( 0, label="Animation for the %d-th curve" % self.file_index )
        # can't post(event.x_root, event.y_root) for matplotlib events
        self.popup_menu.post( int( x + event.x + 10 ), int( y + h_ - event.y + 30 ) )

    def update_cursor( self, cursor ):
        for w in [self.canvas.mpl_canvas_widget]:
            w.config( cursor=cursor )

    def do_animation( self ):
        i = self.file_index
        print( 'do_animation', i )
        self.update_cursor( 'wait' )

        from molass_legacy.KekLib.NumpyUtils             import np_loadtxt
        from SimpleGuinier          import SimpleGuinier
        from molass_legacy.AutorgKekAdapter       import AutorgKekAdapter
        from SimpleGuinierAnimation import SimpleGuinierAnimation
        from CanvasDialog           import CanvasDialog

        if self.averaged_intensity_array is None:
            if USE_SAVED_DATA:
                self.loader.wait_for_xray_loading()
                self.averaged_intensity_array = self.loader.xray_array
                self.datafiles = self.loader.datafiles
            else:
                self.serial_data.wait_until_ready()
                self.serial_data.apply_data_reduction()
                num_curves_averaged = get_setting( 'num_curves_averaged' )
                self.averaged_intensity_array, _, _ = self.serial_data.get_averaged_data( num_curves_averaged )
                self.datafiles  = self.serial_data.datafiles

        if USE_SAVED_DATA:
            averaged_text = ''
        else:
            averaged_text = ' (averaged)'

        print( 'self.num_sg_results=', self.num_sg_results )
        print( 'len( self.averaged_intensity_array )=', len( self.averaged_intensity_array ) )
        assert len( self.averaged_intensity_array ) == self.num_sg_results
        if i < self.num_sg_results:
            data = self.averaged_intensity_array[i, : ]
            file = self.datafiles[i]
            file_for_title = file.replace('\\', '/').split('/')[-1] + averaged_text
        else:
            data, file = self.load_zx_data( i - self.num_sg_results )
            file_for_title = '/'.join( file.replace('\\', '/').split('/')[-2:] )
        guinier = SimpleGuinier( data, anim_data=True )
        adapter = AutorgKekAdapter( None, guinier=guinier )
        adapter_result = adapter.run()

        dialog = CanvasDialog( "SimpleGuinier Animation", parent=self )
        anim_iter_max = 300

        def anim_func( fig  ):
            anim = SimpleGuinierAnimation( guinier, title=file_for_title, result_rg=adapter_result.Rg  )
            anim.draw( fig, anim_iter_max=anim_iter_max )
            dialog.mpl_canvas.draw()

        self.update_cursor( '' )

        dialog.show( anim_func, figsize=( 18, 5 ) )

    def load_zx_data( self, k ):
        j = -1
        k_for_name = k
        found = False
        single_peak = len(self.folder_info) == 1
        half_num = NUM_EXTRAS//2
        for p, tuple_ in enumerate(self.folder_info):
            left, peak, right = tuple_
            left_size = peak - left + half_num + 1 
            for _ in range( left_size ):
                j += 1
                if j == k:
                    found = True
                    ad = 'asc' if single_peak else 'asc(%d)' % (p+1)
                    break
            if found:
                break
            k_for_name -= left_size
            right_size = right - peak + half_num + 1
            for _ in range( right_size ):
                j += 1
                if j == k:
                    found = True
                    ad = 'desc' if single_peak else 'desc(%d)' % (p+1)
                    k -= left_size
                    break
            if found:
                break
            k_for_name -= right_size
        if not found:
            return None, None

        folder = self.report_folder + '/extrapolated/%s' % ( ad )
        if not os.path.exists( folder ):
            return  None, None

        folder_glob_text = folder.replace( '/', '\\' ) + '\\extrapolated-%03d*.*' %  k_for_name
        print( 'folder_glob_text=', folder_glob_text )
        files = glob.glob( folder_glob_text )
        file_path = files[0]
        data, _ = np_loadtxt( file_path )
        return data, file_path

    def show_type_info( self ):
        dialog = TypeInfoPanel( self )
        dialog.show( self.type_names )
        


"""
Baseline.UvBaseline.py

This module has been extracted from the Batch.FullBatch.py
to make its on-the-fly debugging easier.
"""

def get_uv_base_params_impl(editor, xyt=None, debug=False):
    """

    ------------------------------------------------------------------
    editor.base_curve_info  DataTreatment   OptimalTrimming UvPreRecog

    editor.baseline_objects ModeledPeaks    UvPreRecog
    editor.baselines        ModeledPeaks    UvPreRecog
    editor.baseline_params  ModeledPeaks    UvPreRecog
    """
    # uv_base_params has been estimated using UV.UvPreRecog.
    # see QuickAnalysis.ModeledPeaks.get_curve_xy_impl
    uv_base_params = editor.baseline_params[0]
    uv_base_params_ = uv_base_params.copy()

    if debug:
        print("len(uv_base_params_)=", len(uv_base_params_))

    if len(uv_base_params_) == 8 or debug:
        if xyt is None:
            uv_x, uv_y = editor.ecurve_info[0:2]
            uv_ty = editor.uv_ty
        else:
            uv_x, uv_y, uv_ty = xyt

        if len(uv_base_params_) == 8:
            integ_scale, _ = editor.uv_base_curve.guess_integral_scale(uv_x, uv_y, uv_base_params_[:-1], ty=uv_ty)
            uv_base_params_[-1] = integ_scale

    if debug:
        import molass_legacy.KekLib.DebugPlot as plt
        uv_basecurve1, uv_initparams1 = editor.base_curve_info
        baselines = editor.baselines
        uv_basecurve2 = editor.baseline_objects[0]
        uv_initparams2 = uv_base_params_
        with plt.Dp():
            fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12,5))
            fig.suptitle("get_uv_base_params_impl")

            ax1.set_title("from DataTreatment")
            ax1.plot(uv_x, uv_y, label="data")
            ax1.plot(uv_x, uv_basecurve1(uv_x, uv_initparams1, uv_ty), label="baseline")
            ax1.legend()

            ax2.set_title("from ModeledPeaks")
            ax2.plot(uv_x, uv_y, label="data")
            ax2.plot(uv_x, baselines[0], label="baseline 1")
            ax2.plot(uv_x, uv_basecurve2(uv_x, uv_initparams2, uv_ty), label="baseline 2")
            ax2.legend()

            fig.tight_layout()
            plt.show()

    return uv_base_params_


# -*- coding: utf-8 -*-
"""

    ファイル名：   ImageSynthesizer.py

    adapted from ExecutionWindow/bin/MainWindowApp.py

    Copyright (c) 2015-2020, SAXS Team, KEK-PF
"""
import re
from molass_legacy.KekLib.OurTkinter             import Tk
import OurMessageBox            as MessageBox

import logging
import time
import ExecutionWindow
import ViewLog
import ThreadsConnector
import ActionWindow

import numpy as np
from PIL                    import Image
from PilatusImageViewer     import PilatusImageViewer
from PilatusImage           import PilatusImage, pixel_rounded_shift
from SAnglerMask            import SAnglerMask
from Preferences            import get_preference
from SynthesizerSettings    import get_setting, get_mask
from Development            import get_devel_info
from molass_legacy.AutorgKek.AppVersion             import synthesizer_version_string

import gettext
_ = gettext.gettext

UNEXPECTED_ERROR_TEST = False

class ImageSynthesizer():
    def __init__(self, window=None ):
        self.window         = window
        self.logger         = logging.getLogger( __name__ )

    def set_setting_info( self, action ):
        self.mask_array     = get_mask().mask_array
        self.action         = action
        self.syn_method     = get_preference( 'syn_method' )
        self.min_ratio      = get_devel_info( 'min_ratio'  )
        self.syn_flags      = get_preference( 'syn_flags'  )
        self.postfix_adj    = get_devel_info( 'postfix_adj' )
        self.postfix_syn    = get_preference( 'postfix_syn' )

    def execute( self, action, exec_array ):

        self.set_setting_info( action )

        num_selected_rows = len( exec_array )
        if num_selected_rows == 0:
            return

        if action < 3:
            if num_selected_rows > 1:
                MessageBox.showinfo( 'First Selection Notice', 'Only the first selection will be shown.' )
            self.show_images( exec_array[0] )
            return

        # what follows is the case when action == 3
        num_selected_samples = len( exec_array )

        s_ = ''
        if num_selected_samples > 1:
            s_ = 's'

        ok = MessageBox.askokcancel( 
                        'Confirmation',
                        'You are making sythesized images for %d sample%s. Ok?' % ( num_selected_samples, s_ )
                        )
        if not ok:
            return

        self.exec_array = exec_array

        self.logger.info(_('Preparing and starting calculations with %s'), synthesizer_version_string())

        title       = _('Execution Thread Log')
        text        = _('Making %d synthesized images from %s to %s' % ( num_selected_rows, exec_array[0][0], exec_array[num_selected_rows - 1][0] ) )
        geometry    = "1200x500"
        wnd     = ActionWindow.ActionWindow(self.window, title, text, geometry )

        conn = ThreadsConnector.ThreadsConnector()
        conn.runInGui(wnd, conn, None, self.exec_syntheses, 'exec')

    def show_images( self, exec_rec, last_exec_params=[] ):
        # print( 'exec_rec=', exec_rec )
        """
            [0]     Sample id           MAG2wk7210
            [1]     i                   1
            [2]     Original Image      MAG2wk7210_0_00000.tif
            [3]     Absolute Position   ['12.87201', '180.49000'],
            [4]     Shifted Image       MAG2wk7210_1_00000.tif
            [5]     Relative Position   ['5', '5']
            [6]     Intensity Ratio     0.99686741082288877
            [7]     Adjusted Image      MAG2wk7210_1_adj.tif
            [8]     Synthesized Image   MAG2wk7210_1_syn.tif
            [9]     fkey                MAG2wk7210_1
        """

        sample_id   = exec_rec[0]
        in_folder   = get_setting( 'in_folder' )
        adj_folder  = get_setting( 'adj_folder' )
        syn_folder  = get_setting( 'syn_folder' )

        exec_rec_array = exec_rec[1]
        # print( 'exec_rec_array=', exec_rec_array )

        exec_params   = []

        if self.action == 1:
            mask_array = None
        else:
            mask_array = self.mask_array

        o_file  = exec_rec_array[0][0]
        if o_file:
            o_path      = '%s/%s' % ( in_folder, o_file )
            o_pim       = PilatusImage( o_path, mask_array )
            oim_array   = o_pim.image_array()
            exec_params.append( [ o_file, oim_array ] )

        i = 0
        for sub_rec in exec_rec_array[1:]:
            i += 1

            s_delta = []
            i_delta = []
            for d in sub_rec[1]:
                ishift_ = int(float(d) * 1000)
                s_delta.append( ishift_ )
                i_delta.append( pixel_rounded_shift( ishift_ ) )

            if self.action == 1:
                a_file      = sub_rec[0]
                if a_file != None:
                    a_path      = '%s/%s' % ( in_folder, a_file )
                    a_pim       = PilatusImage( a_path )
                    a_im_array  = a_pim.image_array()
                else:
                    continue
            else:
                a_file      = '%s_%d%s.tif' % ( sample_id, i, self.postfix_adj )
                s_file      = sub_rec[0]
                if s_file != None:
                    s_path  = '%s/%s' % ( in_folder, s_file )
                    a_pim   = PilatusImage( s_path, self.mask_array, s_delta[0], s_delta[1] )
                    a_im_array  = a_pim.image_array()
                else:
                    a_im_array  = None

            if type(a_im_array) == np.ndarray:
                exec_params.append( [ a_file, a_im_array, i_delta[0], i_delta[1] ] )


            if i == len( exec_rec_array ) - 1:
                z_file  = sub_rec[4]
                if z_file:
                    z_path  = '%s/%s' % ( syn_folder, z_file )
                    z_pim   = PilatusImage( z_path )
                    z_im_array  = z_pim.image_array()
                    exec_params.append( [ z_file, z_im_array ] )

        self.viewer = PilatusImageViewer( self.action, sample_id, exec_params )

    def exec_syntheses( self, connector, progress ):

        num_rows_to_execute = 0
        for exec_rec in self.exec_array:
            num_rows_to_execute += len( exec_rec[1] ) -1
        # print( 'num_rows_to_execute=', num_rows_to_execute )

        progress.set(0, num_rows_to_execute)

        counter_id  = get_preference( 'detection_counter' )

        for exec_rec in self.exec_array:
            """
                exec_rec example:
                    case of 3 chenges:

                    [ 'AgBh008',
                        [
                            ['AgBh008_0_00000.tif', ['0.70000', '-0.35000'], 1.0, None, None],
                            ['AgBh008_1_00000.tif', ['5', '3'],              1.0, None, None],
                            ['AgBh008_2_00000.tif', ['-5', '-3'],            1.0, None, None]
                        ],
                        3   # num_changes
                    ]

                    case of 2 chenges:
                    [ 'SiGe64201', 
                        [
                            ['SiGe64201_0_00000.tif', ['50.28120', '-0.55000'], 1.0, None, None],
                            ['SiGe64201_1_00000.tif', ['-5', '3'], 1.0, None, None]
                        ],
                        3   # num_changes
                    ]

            """

            # print( 'exec_rec=', exec_rec )
            sample_id       = exec_rec[0]

            try:
                if UNEXPECTED_ERROR_TEST:
                    if sample_id == 'test003_2':
                        unexpected_error()

                exec_rec_array  = exec_rec[1]
                num_changes = exec_rec[2]

                org_name    = '%s_0'  % ( sample_id )
                org_file    = re.sub( r'\.\w+$', '', exec_rec_array[0][0] )
                base_rec    = exec_rec_array[0]

                last_fkey = None
                last_result = []
                last_syn_file = None
                i = 0
                for sub_rec in exec_rec_array[1:]:
                    i += 1
                    exec_seq_no = i
                    fkey        = '%s_%d' % ( sample_id, i )

                    if i > 1:
                        org_name    = last_fkey
                        org_file    = last_syn_file
                        base_rec    = last_result

                    isfinal = i == len( exec_rec_array ) - 1

                    sft_name    = fkey
                    sft_file    = re.sub( r'\.\w+$', '', sub_rec[0] )
                    if isfinal:
                        syn_file    = '%s%s'  % ( sample_id, self.postfix_syn )
                    else:
                        syn_file    = '%s%s'  % ( fkey, self.postfix_syn )

                    i_ratio     = sub_rec[2]

                    # if exec_seq_no < 3 and self.syn_flags[ exec_seq_no ]:
                    if (exec_seq_no >= 3 or self.syn_flags[ exec_seq_no ]) and len(exec_rec_array) == num_changes:
                        counter_id = get_preference( 'detection_counter' )
                        if counter_id == 'None':
                            with_phrase = ''
                        else:
                            with_phrase = ' with %s-ratio %.5f' % ( counter_id, i_ratio )
                        self.logger.info (_('Synthesizing %-20s and %-20s into %-20s%s'), org_file, sft_file, syn_file, with_phrase )

                        last_result = self.exec_single_synthesis( sample_id, base_rec, fkey, sub_rec, isfinal )
                        last_syn_file = syn_file
                    else:
                        self.logger.info (_('Skipping     %-20s and %-20s'), org_file, sft_file )

                    progress.tick()
                    connector.ack() # can be ommitted in this program

            except:
                logging.exception( "Unexpected error while processing '%s' in exec_syntheses" % sample_id )

        self.logger.info(_('Done!'))

    def exec_single_synthesis( self, sample_id, base_rec, fkey, sub_rec, isfinal ):

        in_folder   = get_setting( 'in_folder' )
        adj_folder  = get_setting( 'adj_folder' )
        syn_folder  = get_setting( 'syn_folder' )
        adj_output  = get_devel_info( 'adj_output' )

        o_file = base_rec[0]
        if type(o_file) != np.ndarray:
            o_path  = '%s/%s' % ( in_folder, o_file )
            o_pim   = PilatusImage( o_path, self.mask_array )
            o_ext   = o_pim.image.ext
            self.previous_im = o_pim

            if self.syn_method == 'average':
                # ここで画像のサイズが判明しているので、
                # 平均化に必要な行列を初期化する。
                oim_array = o_pim.image_array()

                # 強度集積用の行列と、有効値カウント用の行列を初期化する。
                self.valid_value_sum_array      = np.zeros( oim_array.shape, dtype='f8' )
                self.valid_pixel_counter_array  = np.zeros( oim_array.shape, dtype='i4' )

                # 正画像の有効値の pixel についてのみ、それぞれ加算する。
                valid_cond = oim_array >= 0
                self.valid_pixel_counter_array[ valid_cond ] += 1
                self.valid_value_sum_array[ valid_cond ] += oim_array[ valid_cond ]
        else:
            o_pim   = PilatusImage( o_file, original_image=self.previous_im )
            o_ext   = self.previous_im.image.ext

        s_file  = sub_rec[0]
        s_path  = '%s/%s' % ( in_folder, s_file )
        s_delta = []
        for d in sub_rec[1]:
            s_delta.append( int(float(d) * 1000) )

        # 副画像を調整する。
        i_ratio = sub_rec[2]
        a_ratio = self.min_ratio
        a_pim   = PilatusImage( s_path, self.mask_array, s_delta[0], s_delta[1], i_ratio, a_ratio )

        if adj_output=='YES' and adj_folder:
            a_file  = '%s%s.%s' % ( fkey, self.postfix_adj, o_ext )
            w_path = '%s/%s' % ( adj_folder, a_file )
            # w_image = Image.fromarray( a_pim.image_array() )
            # w_image.save( w_path, tiffinfo = a_pim.tiffinfo )
            a_pim.save(w_path  )

        if self.syn_method == 'cover':
            # result_im_array = o_pim.make_covered_array( a_pim )
            result_im_array = o_pim.fast_make_covered_array( a_pim )
        else:
            # 調整済副画像の有効値の pixel についてのみ、それぞれ加算する。
            a_im_array  = a_pim.image_array()
            valid_cond  = a_im_array >= 0
            self.valid_pixel_counter_array[ valid_cond ] += 1
            self.valid_value_sum_array[ valid_cond ] += a_im_array[ valid_cond ]
            # result_im_array = o_pim.make_average_array( a_pim )
            result_im_array = o_pim.fast_make_average_array( self.valid_pixel_counter_array, self.valid_value_sum_array )

        if isfinal:
            fkey_ = sample_id
        else:
            fkey_ = fkey

        if isfinal or get_devel_info( 'intermediate_results' ) == 'YES':
            z_file  = '%s%s.%s' % ( fkey_, self.postfix_syn, o_ext )
            w_path  = '%s/%s' % ( syn_folder, z_file )
            # w_image = Image.fromarray( result_im_array )
            # w_image.save( w_path, tiffinfo = a_pim.tiffinfo )
            o_pim.image.set_data( result_im_array, force=True )
            o_pim.save( w_path )

        result = [ result_im_array ]

        return result



"""
    run.py

    Copyright (c) 2021, SAXS Team, KEK-PF
"""

if __name__ == '__main__':
    import os
    import sys
    this_dir = os.path.dirname(os.path.abspath(__file__))
    sys.path.append( this_dir + '/..' )
    from Processes.Main import main
    main()



"""
    ATSAS.Crysol.py

    Copyright (c) 2016-2023, SAXS Team, KEK-PF
"""
import os
import re
import subprocess
from molass_legacy.KekLib.BasicUtils import Struct
from .AutoRg import autorg_exe_array

class CrysolExecutor:
    def __init__(self):
        self.exe_path   = None
        if len( autorg_exe_array ) > 0:
            crysol_path = autorg_exe_array[0].replace( 'autorg.exe', 'crysol_30.exe' )
            if os.path.exists( crysol_path ):
                self.exe_path   = crysol_path

    def execute(self, pdb_file, exp_file=None):
        cmd = [self.exe_path, pdb_file]
        if exp_file is not None:
            cmd += [exp_file]
        result = subprocess.run(cmd, capture_output=True)
        if result.returncode != 0:
            print(result.stderr.decode())
        result.check_returncode()

    def move(self, pdb_file, tgt_folder):
        from shutil import move
        pattern = pdb_file.replace('.pdb', '*00*.*')
        for path in glob.glob(pattern):
            print(path)
            move(path, tgt_folder)

rg_re = re.compile(r"\s+(\S+)$")

def get_info_from_crysol_log(path):
    rg = None
    with open(path) as fh:
        for line in fh:
            if line[0:15].find("Rg (Atoms") >= 0:
                m = rg_re.search(line[:-1])
                if m:
                    rg = float(m.group(1))
    return Struct(Rg=rg)

chi_square_re = re.compile(r"\s+(\S+)$")

def get_info_from_crysol_fit_log(path):
    chi_square = None
    with open(path) as fh:
        for line in fh:
            if line[0:15].find("Chi-square") >= 0:
                m = chi_square_re.search(line[:-1])
                if m:
                    chi_square = float(m.group(1))
    return Struct(Chi_square=chi_square)



# coding: utf-8
"""
    DecompProcess.DecompViewer.py

    Copyright (c) 2021, SAXS Team, KEK-PF
"""
from time import sleep
import numpy as np
from molass_legacy.KekLib.OurTkinter import Tk, Dialog, ttk
import matplotlib.pyplot as plt
from matplotlib.backends.backend_tkagg  import FigureCanvasTkAgg
from matplotlib.gridspec import GridSpec
import molass_legacy.KekLib.DebugPlot as dplt
from molass_legacy.KekLib.OurMatplotlib import NavigationToolbar

class DecompViewer(Dialog):
    def __init__(self, parent):
        self.draw_counter = 0
        Dialog.__init__(self, parent, "Decomposition Viewer", visible=False)

    def show(self):
        self._show()

    def body(self, body_frame):   # overrides parent class method

        cframe = Tk.Frame(body_frame)
        cframe.pack()
        bframe = Tk.Frame(body_frame)
        bframe.pack(fill=Tk.X, expand=1)
        self.bframe = bframe
        self.fig = plt.figure( figsize=(21,11) )
        self.mpl_canvas = FigureCanvasTkAgg( self.fig, cframe )
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack(fill=Tk.BOTH, expand=1)
        self.draw()

    def buttonbox(self):
        tframe = Tk.Frame(self.bframe)
        tframe.pack(side=Tk.LEFT)

        self.toolbar = NavigationToolbar(self.mpl_canvas, tframe)
        self.toolbar.update()

        box = Tk.Frame(self.bframe)
        box.pack(side=Tk.RIGHT)

        w = Tk.Button(box, text="OK", width=10, command=self.ok, default=Tk.ACTIVE)
        w.pack(side=Tk.LEFT, padx=5, pady=5)
        w = Tk.Button(box, text="Cancel", width=10, command=self.cancel)
        w.pack(side=Tk.LEFT, padx=5, pady=5)

        self.bind("<Return>", self.ok)
        self.bind("<Escape>", self.cancel)

    def draw(self):
        fig = self.fig

        gs = GridSpec(3, 4)

        ax1 = fig.add_subplot(gs[0,0])
        axt = ax1.twinx()
        axt.grid(False)

        ax2 = fig.add_subplot(gs[0,1], projection='3d')
        ax3 = fig.add_subplot(gs[0,2])
        ax4 = fig.add_subplot(gs[0,3])

        axes = np.empty((2,4), dtype=object)
        for i in range(2):
            for j in range(4):
                axes[i,j] = fig.add_subplot(gs[i+1,j])

        fig.tight_layout()

        self.after(100, lambda: self.draw_subplots(fig, axes))
        self.mpl_canvas.draw()

    def draw_subplots(self, fig, axes):
        print("draw_counter=", self.draw_counter)

        i, j = divmod(self.draw_counter, 4)

        ax = axes[i,j]
        ax.plot(0, 0, 'o', markersize=50)
        fig.canvas.draw()

        self.draw_counter += 1
        if self.draw_counter < 8:
            self.after(1000, lambda: self.draw_subplots(fig, axes))



"""
    SecTheory.EdmSpike.py

    EDM - Equilibrium Dispersive Model - Spike

    Copyright (c) 2023, SAXS Team, KEK-PF
"""
import numpy as np
import io
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
from matplotlib.patches import Rectangle
import seaborn
seaborn.set()

import molass_legacy.KekLib.DebugPlot as plt
from DataUtils import get_in_folder
from .Edm import guess_multiple_edms

def guess_edm_initial_params(x, y, return_edms=False):

    slice_ = slice(200, 300)
    edms_ = guess_multiple_edms(x[slice_], y[slice_], 1)
    edm1 = edms_[0]

    slice_ = slice(300, 350)
    edms_ = guess_multiple_edms(x[slice_], y[slice_], 1)
    edm2 = edms_[0]

    y_ = y - edm1(x) - edm2(x)
    y_[y_ < 0] = 0

    slice_ = slice(100, 200)
    edms_ = guess_multiple_edms(x[slice_], y_[slice_], 1)
    edm0 = edms_[0]
    edms = [edm0, edm1, edm2]

    if return_edms:
        return edms

    xr_params = []
    for edm in edms:
        xr_params.append(edm.get_comp_params())

    return np.array(xr_params)

def spike(in_folder, sd, num_components):
    from .Edm import edm_func

    print("spike_impl", in_folder, num_components)

    D, E, qv, xr_curve = sd.get_xr_data_separate_ly()

    x = xr_curve.x
    y = xr_curve.y
    params = guess_edm_initial_params(x, y, return_edms=False)

    u = 0.5

    with plt.Dp():
        fig, ax = plt.subplots()
        ax.set_title("EDM initial fitting to %s" % get_in_folder(in_folder), fontsize=20)
        ax.plot(x, y, color="orange")

        ty = np.zeros(len(y))
        for k, p in enumerate(params):
            print([k], *p)
            cy = edm_func(x, u, *p)
            ty += cy
            ax.plot(x, cy, ":", label="component-%d" % (k+1))

        ax.plot(x, ty, ":", color="red", lw=2, label="component total")

        ax.legend(loc="upper left", fontsize=16)

        fig.tight_layout()
        plt.show()
